# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module defines the TreeBuilder, which is used to build() a tree structure
from a text, using a root lexicon.

Using the rebuild() method, TreeBuilder is also capable of regenerating only
part of an existing tree, e.g. when part of a long text is modified through a
text editor. It is smart enough to recognize whether existing tokens before and
after the modified region can be reused or not, and it reuses tokens as much as
possible.

BackgroundTreeBuilder can be used to tokenize a text into a tree in a background
thread.

"""


import sys
import contextlib
import itertools
import threading

from livelex.action import DynamicAction
from livelex.target import DynamicTarget
from livelex.tree import Context, Token, _GroupToken


class TreeBuilder:
    """Build a tree directly from parsing the text.

    The root node of the tree is in the `root` instance attribute.

    After calling build() or rebuild(), three instance variables are set:

        start, end:
            indicate the region the tokens were changed. After build(), start
            is always 0 and end = len(text), but after rebuild(), these values
            indicate the range that was actually re-tokenized.

        lexicons:
            the list of open lexicons (excluding the root lexicon) at the end
            of the document. This way you can see in which lexicon parsing
            ended.

            If a tree was rebuilt, and old tail tokens were reused, the
            lexicons variable is not set, meaning that the old value is still
            valid. If the TreeBuilder was not used before, lexicons is then
            None.

    No other variables or state are kept, so if you don't need the above
    information anymore, you can throw away the TreeBuilder after use.

    """
    changes = None

    start = 0
    end = 0
    lexicons = None

    def __init__(self, root_lexicon=None):
        self.root = Context(root_lexicon, None)

    def tree(self, text):
        """Convenience method returning the tree with all tokens."""
        self.build(text)
        return self.root

    def build(self, text):
        """Tokenize the full text.

        Sets three instance variables start, end, lexicons). Start and end
        are always 0 and len(text), respectively. lexicons is a list of the
        lexicons that were not closed at the end of the text. (If the parser
        ended in the root context, the list is empty.)

        """
        self.rebuild(text, 0, 0, len(text))

    def rebuild(self, text, start, removed, added):
        """Tokenize the modified part of the text again and update the tree.

        Sets, just like build(), three instance variables start, end, lexicons,
        describing the region in the thext the tokens were changed. This range
        can be larger than (start, start + added).

        The text is the new text; start is the position where characters were
        removed and others added. The removed and added arguments are integers,
        describing how many characters were removed and added.

        This method finds the place we can start parsing again, and when the
        end of the modified region is reached, automatically recognizes when
        the rest of the tokens can be reused. When old tokens at the end are
        reused, the lexicons instance variable is not reset, the existing
        value is still relevant in that case.

        """
        if not self.root.lexicon:
            self.root.clear()
            self.start, self.end = start, start + added
            return

        # manage end, and record if there is text after the modified part (tail)
        end = start + removed
        tail = start + added < len(text)

        # we may be able to use existing tokens for the start if start > 0
        head = start > 0

        # record the position change for tail tokens that maybe are reused
        offset = added - removed

        # If there remains text after the modified part,
        # we try to reuse the old tokens
        if tail:
            # find the first token after the modified part
            end_token = self.root.find_token_after(end)
            if end_token:
                # make a subtree structure starting with this end_token
                tail_tree = end_token.split()
                tail_gen = ((t, t.pos + offset)
                    for t in tail_tree.tokens()
                        if not t.group or (t.group and t is t.group[0]))
                # store the new position the first tail token would get
                for tail_token, tail_pos in tail_gen:
                    break
                else:
                    tail = False
            else:
                tail = False

        lowest_start = sys.maxsize
        restart = True
        while restart:
            restart = False

            # find the last token before the modified part, we will start parsing
            # before that token. If there are no tokens, we just start at 0.
            # At least go back to just before a newline, if possible.
            if head:
                i = text.rfind('\n', 0, start)
                start_token = self.root.find_token_before(i) if i > -1 else None
                if not start_token:
                    start_token = self.root.find_token_before(start)
                    if start_token:
                        # go back some more tokens, you never know a longer match
                        # could be made. In very particular cases a longer token
                        # could be found. (That's why we tried to go back to a
                        # newline.)
                        gen = start_token.backward()
                        for start_token in itertools.islice(gen, 9):
                            pass
                        start_token = next(gen, None)
                if start_token:
                    # don't start in the middle of a group, as they originate from
                    # one single regexp match
                    if start_token.group:
                        start_token = start_token.group[0]

                    # make a short list of tokens from the start_token to the place
                    # we want to parse. We copy them because some might get moved to
                    # the tail self.root. If they were not changed, we can adjust the
                    # modified region.
                    start_tokens = [start_token.copy()]
                    for t in start_token.forward():
                        start_tokens.append(t.copy())
                        if t.end > start:
                            break
                    start_token_index = 0
                else:
                    head = False

            if head:
                # remove the start token and all tokens to the right
                pos = start = start_token.pos
                context = start_token.parent
                start_token.cut()
            else:
                pos = start = lowest_start = 0
                context = self.root
                context.clear()

            # start parsing
            done = False
            while not done:
                for pos, tokens, target in self.parse_context(context, text, pos):
                    if tokens:
                        if head:
                            # move start if the tokens before start didn't change
                            if (start_token_index + len(tokens) <= len(start_tokens) and
                                all(new.equals(old)
                                    for old, new in zip(start_tokens[start_token_index:], tokens))):
                                start = pos
                                start_token_index += len(tokens)
                            else:
                                start = tokens[0].pos
                                head = False    # stop looking further
                        if tail:
                            if tokens[0].pos > tail_pos:
                                for tail_token, tail_pos in tail_gen:
                                    if tail_pos >= tokens[0].pos:
                                        break
                                else:
                                    tail = False
                            if (tokens[0].pos == tail_pos
                                    and tokens[0].state_matches(tail_token)):
                                # we can attach the tail here.
                                if offset:
                                    # adjust the pos of the old tail tokens.
                                    # We don't use tail_token.forward() because
                                    # it uses parent_index() which depends on sorted
                                    # pos values
                                    tail_token.cut_left()
                                    for t in tail_tree.tokens():
                                        t.pos += offset
                                # add the old tokens to the current context
                                tail_token.join(context)
                                end = tail_pos
                                done = True
                                break
                        context.extend(tokens)
                        # we check for new changes here, so we always have tokens
                        # in the current context
                        if self.changes:
                            c = self.get_changes()
                            if c:
                                # break out and adjust the current tokenizing process
                                text = c.text
                                start = c.position
                                head = start > 0
                                if c.root_lexicon != False:
                                    self.root.lexicon = c.root_lexicon
                                    head = tail = False
                                elif tail:
                                    # reuse old tail?
                                    new_tail_pos = start + c.added
                                    if new_tail_pos >= len(text):
                                        tail = False
                                    else:
                                        offset += c.added - c.removed
                                        tail_pos += offset
                                        if new_tail_pos > tail_pos:
                                            for tail_token, tail_pos in tail_gen:
                                                if tail_pos >= new_tail_pos:
                                                    break
                                            else:
                                                tail = False
                                restart = done = True
                                break # restart at new start position
                    if target:
                        context = self.update_context(context, target)
                        break # continue with new context
                else:
                    end = len(text)
                    self.unwind(context)
                    break
        self.start, self.end = min(start, lowest_start), end

    def unwind(self, context):
        """Recursively remove the context from its parent if empty.

        Leaves the list of lexicons that were left open in the `lexicons`
        attribute. When parsing ended in the root context, that list is empty.

        """
        self.lexicons = []
        while context.parent:
            self.lexicons.append(context.lexicon)
            if not context:
                del context.parent[-1]
            context = context.parent

    def parse_context(self, context, text, pos):
        """Yield Token instances as long as we are in the current context."""
        for pos, txt, match, action, *target in context.lexicon.parse(text, pos):
            if txt:
                if isinstance(action, DynamicAction):
                    tokens = tuple(action.filter_actions(self, pos, txt, match))
                    if len(tokens) == 1:
                        tokens = Token(context, *tokens[0]),
                    else:
                        tokens = tuple(_GroupToken(context, *t) for t in tokens)
                        for t in tokens:
                            t.group = tokens
                else:
                    tokens = Token(context, pos, txt, action),
            else:
                tokens = ()
            if target and isinstance(target[0], DynamicTarget):
                target = target[0].target(match)
            yield pos + len(txt), tokens, target

    def update_context(self, context, target):
        """Move to another context depending on target."""
        for t in target:
            if isinstance(t, int):
                for pop in range(t, 0):
                    if context.parent:
                        if not context:
                            del context.parent[-1]
                        context = context.parent
                    else:
                        break
                for push in range(0, t):
                    context = Context(context.lexicon, context)
                    context.parent.append(context)
            else:
                context = Context(t, context)
                context.parent.append(context)
        return context

    def filter_actions(self, action, pos, txt, match):
        """Handle filtering via DynamicAction instances."""
        if isinstance(action, DynamicAction):
            yield from action.filter_actions(self, pos, txt, match)
        elif txt:
            yield pos, txt, action


class BackgroundTreeBuilder(TreeBuilder):
    """A TreeBuilder that can tokenize a text in a background thread.

    BackgroundTreeBuilder supports tokenizing in a background thread. You
    must specify changes to the document using the Changes object returned by
    the change() context manager method:

        with builder.change() as c:
            c.change_text("new text", position, removed, added)

    Tokenizing then starts in the background. During tokenizing it is still
    possible to add new changes using change(). You can add callbacks to the
    `updated_callbacks` attribute (using add_build_updated_callback()) that are
    called everytime the whole document is tokenized.

    You can also add callbacks to the `finished_callbacks` attribute, using
    add_finished_callback(); those are called once when all pending changes are
    processed and then forgotten again.

    To be sure you get a complete tree, call get_root().

    """
    def __init__(self, root_lexicon=None):
        super().__init__(root_lexicon)
        self.job = None
        self.changes = None
        self.lock = threading.Lock()
        self.updated_callbacks = []
        self.finished_callbacks = []

    @contextlib.contextmanager
    def change(self):
        """Return a Changes object and start a context to add changes."""
        with self.lock:
            c = self.changes or Changes()
            yield c
            if c.has_changes():
                self.changes = c
                self.start_processing()

    def start_processing(self):
        """Start a background job if needed."""
        if not self.job:
            self.job = threading.Thread(target=self.process_and_finish)
            self.job.start()

    def process_and_finish(self):
        """Call process_changes() and then finish_processing()."""
        self.process_changes()
        self.finish_processing()

    def process_changes(self):
        """Process changes as long as they are added. Called in the background."""
        c = self.get_changes()
        start = -1
        end = -1
        while c and c.has_changes():
            if c.root_lexicon != False and c.root_lexicon != self.root.lexicon:
                self.root.lexicon = c.root_lexicon
                self.build(c.text)
            else:
                self.rebuild(c.text, c.position, c.removed, c.added)
            start = self.start if start == -1 else min(start, self.start)
            end = self.end if end == -1 else max(c.new_position(end), self.end)
            c = self.get_changes()
        if start != -1:
            self.start = start
        if end != -1:
            self.end = end

    def finish_processing(self):
        """Called when process_changes() quits. Calls oneshot callbacks.

        In GUI applications, this method should run when the job has finished,
        in the GUI thread.

        """
        with self.lock:
            self.job = None
        # this happens when still changes were added.
        c = self.changes
        if c and c.has_changes():
            start, end = self.start, self.end
            self.process_changes()
            self.start = min(start, self.start)
            self.end = max(c.new_position(end), self.end)
        self.build_updated()
        while self.finished_callbacks:
            callback, args, kwargs = self.finished_callbacks.pop()
            callback(*args, **kwargs)

    def get_changes(self):
        """Get the changes once. To be used from within the thread."""
        with self.lock:
            c, self.changes = self.changes, None
            return c

    def wait(self):
        """Wait for completion if a background job is running."""
        job = self.job
        if job:
            job.join()

    def get_root(self, wait=False, callback=None, args=None, kwargs=None):
        """Get the root element of the completed tree.

        If wait is True, this call blocks until tokenizing is done, and the
        full tree is returned. If wait is False, None is returned if the tree
        is still busy being built.

        If a callback is given and tokenizing is still busy, that callback is
        called once when tokenizing is ready. If given, args and kwargs are the
        arguments the callback is called with, defaulting to () and {},
        respectively.

        Note that, for the lifetime of a TreeBuilder, the root element is always
        the same. The root element is also accessible in the `root` attribute.
        But using this method you can be sure that you are dealing with a
        complete and fully intact tree.

        """
        with self.lock:
            job = self.job
            if not job:
                return self.root
            if callback:
                self.add_finished_callback(callback, args, kwargs)
        if wait:
            self.wait()
            return self.root

    def build_updated(self):
        """Called when a build() or rebuild() finished and the tree is complete.

        The default implementation calls all callbacks in the
        `updated_callbacks` attribute, with the (start, end) arguments. (The
        same values are also accessible in the `start` and `end` attributes.)

        """
        for cb in self.updated_callbacks:
            cb(self.start, self.end)

    def add_build_updated_callback(self, callback):
        """Add a callback to be called when the whole text is tokenized.

        The callback is called with two arguments (start, end) denoting
        the range in the text that was tokenized again.

        """
        if callback not in self.updated_callbacks:
            self.updated_callbacks.append(callback)

    def remove_build_updated_callback(self, callback):
        """Remove a previously registered callback to be called when the whole text is tokenized."""
        if callback in self.updated_callbacks:
            self.updated_callbacks.remove(callback)

    def add_finished_callback(self, callback, args=None, kwargs=None):
        """Add a callback to be called when tokenizing finishes.

        This callback will be called once, directly after being called
        it will be forgotten.

        """
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        cb = (callback, args, kwargs)
        if cb not in self.finished_callbacks:
            self.finished_callbacks.append(cb)

    def remove_finished_callback(self, callback, args=None, kwargs=None):
        """Remove a callback that was registered to be called when tokenizing finishes."""
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        cb = (callback, args, kwargs)
        if cb in self.finished_callbacks:
            self.finished_callbacks.remove(cb)


class Changes:
    """Store changes that have to be made to a tree.

    This object is used through TreeBuilder.change().
    Calling change_contents() merges new changes with the existing changes.
    Calling change_root_lexicon() stores a root lexicon change.
    On init and clear() the changes are reset.

    """
    __slots__ = "root_lexicon", "text", "position", "added", "removed"

    def __init__(self):
        self.root_lexicon = False   # meaning no change is requested
        self.text = ""
        self.position = -1          # meaning no text is altered
        self.removed = 0
        self.added = 0

    def __repr__(self):
        changes = []
        if self.root_lexicon != False:
            changes.append("root_lexicon: {}".format(self.root_lexicon))
        if self.position != -1:
            changes.append("text: {} -{} +{}".format(self.position, self.removed, self.added))
        if not changes:
            changes.append("(no changes)")
        return "<Changes {}>".format(', '.join(changes))

    def change_contents(self, text, position=0, removed=None, added=None):
        """Merge new change with existing changes.

        If added and removed are not given, all text after position is
        considered to be replaced.

        """
        if removed is None:
            removed = len(self.text) - position
        if added is None:
            added = len(text) - position
        self.text = text
        if self.position == -1:
            # there were no previous changes
            self.position = position
            self.removed = removed
            self.added = added
            return
        # determine the offset for removed and added
        if position + removed < self.position:
            offset = self.position - position - removed
        elif position > self.position + self.added:
            offset = position - self.position - self.added
        else:
            offset = 0
        # determine which part of removed falls inside existing changes
        start = max(position, self.position)
        end = min(position + removed, self.position + self.added)
        offset -= max(0, end - start)
        # set the new values
        self.position = min(self.position, position)
        self.removed += removed + offset
        self.added += added + offset

    def change_root_lexicon(self, text, root_lexicon):
        """Store a root lexicon change."""
        self.text = text
        self.root_lexicon = root_lexicon

    def has_changes(self):
        """Return True when there are actually changes."""
        return self.position != -1 or self.root_lexicon != False

    def new_position(self, pos):
        """Return how the current changes would affect an older position."""
        if pos < self.position:
            return pos
        elif pos < self.position + self.removed:
            return self.position + self.added
        return pos - self.removed + self.added




# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re


from livelex import *

RE_SCHEME_RIGHT_BOUND = r"(?=$|[)\s])"
RE_SCHEME_FRACTION = r"-?\d+/\d+" + RE_SCHEME_RIGHT_BOUND
RE_SCHEME_FLOAT = r"-?((\d+(\.\d*)|\.\d+)(E\d+)?)" + RE_SCHEME_RIGHT_BOUND
RE_SCHEME_NUMBER = (
    r"("
    r"-?\d+|"
    r"#(b[0-1]+|o[0-7]+|x[0-9a-fA-F]+)|"
    r"[-+]inf.0|[-+]?nan.0"
    r")" + RE_SCHEME_RIGHT_BOUND
)

class Scheme(Language):
    @lexicon
    def root(cls):
        yield from cls.common()
    
    @classmethod
    def common(cls):
        yield r"['`,]", Delimiter.Scheme.Quote
        yield r"\(", Delimiter.OpenParen, cls.list
        yield r"#\(", Delimiter.OpenVector, cls.vector
        yield r'"', String, cls.string
        yield r';', Comment, cls.singleline_comment
        yield r'#!', Comment, cls.multiline_comment
        yield RE_SCHEME_FRACTION, Number.Fraction
        yield RE_SCHEME_FLOAT, Number.Float
        yield RE_SCHEME_NUMBER, Number
        yield r"\.(?!\S)", Delimiter.Dot
        yield r"#[tf]\b", Boolean
        yield r"#\\([a-z]+|.)", Char
        yield r'[^()"{}\s]+', cls.get_word_action()

    @lexicon
    def list(cls):
        yield r"\)", Delimiter.CloseParen, -1
        yield from cls.common()
        
    @lexicon
    def vector(cls):
        yield r"\)", Delimiter.CloseVector, -1
        yield from cls.common()
        
    @classmethod
    def get_word_action(cls):
        """Return a dynamic action that is chosen based on the text."""
        def test(text):
            from . import scheme_words
            return text in scheme_words.keywords
        return bytext(test, Name, Keyword)

        
    # -------------- String ---------------------
    @lexicon
    def string(cls):
        yield r'"', String, -1
        yield from cls.string_common()

    @classmethod
    def string_common(cls):
        yield r'\\[\\"]', String.Escape
        yield default_action, String

    # -------------- Comment ---------------------
    @lexicon
    def multiline_comment(cls):
        yield r'!#', Comment, -1
        yield from cls.comment_common()

    @lexicon(re_flags=re.MULTILINE)
    def singleline_comment(cls):
        yield from cls.comment_common()
        yield r'$', Comment, -1

    @classmethod
    def comment_common(cls):
        yield default_action, Comment


class SchemeLily(Scheme):
    """Scheme used with LilyPond."""
    @lexicon
    def one_arg(cls):
        """Pick one thing and pop back."""
        yield r"['`,]", Delimiter.Scheme.Quote
        yield r"\(", Delimiter.OpenParen, -1, cls.list
        yield r"#\(", Delimiter.OpenVector, -1, cls.vector
        yield r'"', String, -1, cls.string
        yield r';', Comment, -1, cls.singleline_comment
        yield r'#!', Comment, -1, cls.multiline_comment
        yield RE_SCHEME_FRACTION, Number.Fraction, -1
        yield RE_SCHEME_FLOAT, Number.Float, -1
        yield RE_SCHEME_NUMBER, Number, -1
        yield r"#[tf]\b", Boolean, -1
        yield r"#\\([a-z]+|.)", Char, -1
        yield r'[^()"{}\s]+', cls.get_word_action(), -1
        
    @classmethod
    def common(cls):
        yield from super().common()
        from . import lilypond
        yield r"#{", Delimiter.LilyPond, lilypond.LilyPond.schemelily




# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
A Lexicon groups rules to match.

A Lexicon is created by decorating a function yielding rules with the
`@lexicon` decorator. A Lexicon acts as a descriptor; when accessed for
the first time via a Language class, a BoundLexicon for that class is
created and cached, and returned each time that attribute is accessed.

The BoundLexicon can parse text according to the rules. When parsing for the
first time, the rules-function is run with the language class as argument, and
the rules it creates are cached.

This makes it possible to inherit from a Language class and only re-implement
some lexicons, the others keep working as in the base class.

"""

import re
import threading

import livelex.action
import livelex.pattern
import livelex.regex


class Lexicon:
    """A Lexicon consists of a set of pattern rules a text is scanned for."""
    __slots__ = ('rules_func', 'lexicons', '_lock', 're_flags')

    def __init__(self, rules_func,
                       re_flags=0,
        ):
        """Initializes with the rules function.

        The rules function accepts the Language class as argument, and yields
        the pattern, action, target, ... tuples.

        """
        self.rules_func = rules_func
        self.re_flags = re_flags
        self.lexicons = {}
        self._lock = threading.Lock()

    def __get__(self, instance, owner):
        """Called when accessed as a descriptor, via the Language class."""
        if instance:
            raise RuntimeError('Language should never be instantiated')
        try:
            return self.lexicons[owner]
        except KeyError:
            # prevent instantiating the same BoundLexicon multiple times
            with self._lock:
                try:
                    lexicon = self.lexicons[owner]
                except KeyError:
                    lexicon = self.lexicons[owner] = BoundLexicon(self, owner)
                return lexicon


class BoundLexicon:
    """A Bound Lexicon is tied to a particular class.

    This makes it possible to inherit from a Language class and change
    only some Lexicons.

    Call BoundLexicon.parse(text, pos) to do the actual parsing work.
    This function is created as soon as it is called for the first time.

    """
    __slots__ = ('lexicon', 'language', 'parse', '_lock')

    def __init__(self, lexicon, language):
        self.lexicon = lexicon
        self.language = language
        # lock is used once when creating the parse() instance function
        self._lock = threading.Lock()

    def __call__(self):
        """Call the original function, yielding the rules."""
        return self.lexicon.rules_func(self.language) or ()

    def __repr__(self):
        return self.name()

    def name(self):
        """Return the 'Language.lexicon' name of this bound lexicon."""
        return '.'.join((self.language.__name__, self.lexicon.rules_func.__name__))

    def __getattr__(self, name):
        """Implemented to create the parse() function when called for the first time.

        The function is created by _get_parser_func() and is set as instance
        variable, so __getattr__ is not called again.

        """
        try:
            lock = object.__getattribute__(self, "_lock")
        except AttributeError:
            pass # the lock was already deleted, meaning parse is in place already
        else:
            with lock:
                if name == "parse":
                    try:
                        return object.__getattribute__(self, name)
                    except AttributeError:
                        self.parse = self._get_parser_func()
                    del self._lock
                    return self.parse
        return object.__getattribute__(self, name)

    @property
    def re_flags(self):
        """The re_flags set on instantiation."""
        return self.lexicon.re_flags

    def _get_parser_func(self):
        """Compile the pattern rules and return a parse(text, pos) func."""
        patterns = []
        action_targets = []
        default_action = None
        default_target = None
        # make lists of pattern, action and possible targets
        for pattern, action, *target in self():
            if pattern is livelex.default_action:
                default_action = action
            elif pattern is livelex.default_target:
                default_target = action, *target
            else:
                if isinstance(pattern, livelex.pattern.Pattern):
                    pattern = pattern.build()
                patterns.append(pattern)
                action_targets.append((action, *target))

        if not patterns:
            if default_action:
                def parse(text, pos):
                    yield pos, text[pos:], None, default_action
            elif default_target:
                def parse(text, pos):
                    yield (pos, "", None, None, *default_target)
            else:
                # just quits parsing
                def parse(text, pos):
                    yield from ()
            return parse

        # if there is only one pattern, and no dynamic action or target,
        # see if the pattern is simple enough to just use str.find
        if (len(patterns) == 1
            and not isinstance(action_targets[0][0], livelex.action.DynamicAction)
            and not any(isinstance(target, livelex.target.DynamicTarget)
                            for target in action_targets[0][1:])):
            needle = livelex.regex.to_string(patterns[0])
            if needle:
                l= len(needle)
                action_target = action_targets[0]
                if default_action:
                    def parse(text, pos):
                        """Parse text, using a default action for unknown text."""
                        while True:
                            i = text.find(needle, pos)
                            if i > pos:
                                yield pos, text[pos:i], None, default_action
                            elif i == -1:
                                break
                            yield (i, needle, None, *action_target)
                            pos = i + l
                        if pos < len(text):
                            yield pos, text[pos:], None, default_action
                elif default_target:
                    def parse(text, pos):
                        """Parse text, stopping with the default target at unknown text."""
                        while needle == text[pos:pos+l]:
                            yield (pos, needle, None, *action_target)
                            pos += l
                        yield (pos, "", None, None, *default_target)
                else:
                    def parse(text, pos):
                        """Parse text, skipping unknown text."""
                        while True:
                            i = text.find(needle, pos)
                            if i == -1:
                                break
                            yield (i, needle, None, *action_target)
                            pos = i + l
                return parse
        # compile the regexp for all patterns
        rx = re.compile("|".join("(?P<g_{0}>{1})".format(i, pattern)
            for i, pattern in enumerate(patterns)), self.re_flags)
        # make a fast mapping list from matchObj.lastindex to the targets
        indices = sorted(v for k, v in rx.groupindex.items() if k.startswith('g_'))
        index = [None] * (indices[-1] + 1)
        for i, action_target in zip(indices, action_targets):
            index[i] = action_target

        if default_action:
            def parse(text, pos):
                """Parse text, using a default action for unknown text."""
                for m in rx.finditer(text, pos):
                    if m.start() > pos:
                        yield pos, text[pos:m.start()], None, default_action
                    yield (m.start(), m.group(), m, *index[m.lastindex])
                    pos = m.end()
                if pos < len(text):
                    yield (pos, text[pos:], None, default_action)
        elif default_target:
            def parse(text, pos):
                """Parse text, stopping with the default target at unknown text."""
                while True:
                    m = rx.match(text, pos)
                    if m:
                        yield (pos, m.group(), m, *index[m.lastindex])
                        pos = m.end()
                    else:
                        yield (pos, "", None, None, *default_target)
                        break
        else:
            def parse(text, pos):
                """Parse text, skipping unknown text."""
                for m in rx.finditer(text, pos):
                    yield (m.start(), m.group(), m, *index[m.lastindex])
        return parse





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The action module defines two distinct base classes for actions.

StandardAction is the base class for fixed standard actions, and DynamicAction
is the base class for objects that yield actions based on the matched text.

In a livelex rule, an action may be any object, e.g. a method, number or string
literal. So it is not necessary that you use the provided standard actions. The
dynamic actions can also be used with your own actions, and you can also
inherit from DynamicAction if you want.

These standard actions are intended to help create a uniform handling of
actions across multiple languages and lexicons.


StandardAction
==============

A StandardAction is a singleton object. Acessing an attribute (without
underscore) creates that attribute as a new instance, with the current instance
as _parent.

This way a new action type can be created that shares its parent with other
types, a concept borrowed from pygments. An example:

    >>> Comment = StandardAction("Comment")
    >>> Literal = StandardAction("Literal")
    >>> String = Literal.String
    >>> String.DoubleQuotedString
    Literal.String.DoubleQuotedString
    >>> String.SingleQuotedString
    Literal.String.SingleQuotedString

StandardAction instances support iteration and membership methods.
Iteration yields the instance ifself and then the parents:

    >>> for i in String.DoubleQuoted:
    ...     print(i)
    ...
    Literal.String.DoubleQuoted
    Literal.String
    Literal

And the `in` operator returns True when a standard action belongs to another
one, i.e. the other one is one of the ancestors of the current action:

    >>> String.DoubleQuotedString in String
    True
    >>> Literal in String
    False

Finally, the `&` operator returns the common ancestor, if any.

This module defines the following pre-defined standard actions:

    Whitespace = action.StandardAction("Whitespace")
    Text = action.StandardAction("Text")

    Escape = action.StandardAction("Escape")
    Keyword = action.StandardAction("Keyword")
    Name = action.StandardAction("Name")
    Literal = action.StandardAction("Literal")
    Delimiter = action.StandardAction("Delimiter")
    Comment = action.StandardAction("Comment")
    Error = action.StandardAction("Error")

    Verbatim= Literal.Verbatim
    String = Literal.String
    Number = Literal.Number
    Builtin = Name.Builtin
    Variable = Name.Variable

(This list may be out of date, see __init__.py for the exact list.)


DynamicAction
=============

If an instance of DynamicAction is encountered in a rule, its filter_actions()
method is called to yield a (pos, text action) tuple. Normally the
filter_actions() method simply calls back the filter_actions() method of the
lexer with the new action, which could again be an Action instance.

Nesting is possible in most cases, only some actions require the match object
to be present; and such actions can't be used as default action, or inside
subgroup_actions.

A DynamicAction object always holds all actions it is able to return in the
actions attribute. This is done so that it is possible to know all actions a
Language can generate beforehand, and e.g. translate all the actions in a
Language to other objects, which could even be methods or functions.

SubgroupAction
--------------

A SubgroupAction looks at subgroups in the regular expression match and
returns the same amount of tokens as there are subgroups, using the specified
action for every subgroup.

For example, the rule:

    "(0x)([0-9a-f]+)", SubgroupAction(Number.Prefix, Number.Hexadecimal)

    yields two tokens in case of a match, one for "0x" and the other for the
    other group of the match.

TextAction and MatchAction
--------------------------

Those expect a preficate function as the first argument, and one or more
actions as further arguments.

The predicate is run if the rule matches, for TextAction with the text
as argument, and for MatchAction with the match object as argument. The
return value is the index of the action to pick. False and True count as
0 and 1 respectively.

A _SkipAction() is stored in the module variable `skip` and causes the rule
to silently ignore the matched text.

"""


class StandardAction:
    """Factory for standard action singletons."""
    _parent = None
    def __init__(self, name):
        self._name = name

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError("{} has no attribute {}".format(self, repr(name)))
        new = type(self)(name)
        new._parent = self
        setattr(self, name, new)
        return new

    def __repr__(self):
        return ".".join(reversed([n._name for n in self]))

    def __iter__(self):
        node = self
        while node:
            yield node
            node = node._parent

    def __contains__(self, other):
        return any(t is self for t in other)

    def __and__(self, other):
        ancestors = frozenset(other)
        for t in self:
            if t in ancestors:
                return t

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class DynamicAction:
    """Base class for dynamic action objects.

    All actions a DynamicAction object can yield are in the actions attribute.

    """
    def __init__(self, *actions):
        self.actions = actions

    def filter_actions(self, lexer, pos, text, match):
        raise NotImplementedError


class SubgroupAction(DynamicAction):
    """Yield actions from subgroups in a match.

    There should be the same number of subgroups in the regular expression as
    there are action attributes given to __init__().

    """
    def filter_actions(self, lexer, pos, text, match):
        for i, action in enumerate(self.actions, match.lastindex + 1):
            yield from lexer.filter_actions(action, match.start(i), match.group(i), None)


class PredicateAction(DynamicAction):
    """Base class expecting a predicate function and actions."""
    def __init__(self, predicate, *actions):
        self.predicate = predicate
        super().__init__(*actions)

    def index(self, text, match):
        raise NotImplementedError

    def filter_actions(self, lexer, pos, text, match):
        index = self.index(text, match)
        action = self.actions[index]
        yield from lexer.filter_actions(action, pos, text, match)


class MatchAction(PredicateAction):
    """Expects a function as argument that is called with the match object.

    The function should return the index indicating the action to return.
    The function may also return True or False, which are regarded as 1 or 0,
    respectively.

    """
    def index(self, text, match):
        return self.predicate(match)


class TextAction(PredicateAction):
    """Expects a function as argument that is called with the matched text.

    The function should return the index indicating the action to return.
    The function may also return True or False, which are regarded as 1 or 0,
    respectively.

    """
    def index(self, text, match):
        return self.predicate(text)


class SkipAction(DynamicAction):
    """A DynamicAction that yields nothing."""
    def filter_actions(self, lexer, pos, text, match):
        yield from ()




# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The target module defines targets that depend on the match object from the
lexer.

Instead of zero or more targets (which may be lexicons or integers), one
DynamicTarget instance can be given as DynamicTarget in a rule. Its target()
method is then called to return the desired tuple of targets.

"""


from livelex.lexicon import BoundLexicon


class DynamicTarget:
    """Base class for DynamicTarget objects, that choose a target at runtime.

    The possible targets are in the targets attribute.
    An index() function returns the index of the target to choose.

    If you specify a single integer or lexicon as a target, it is automatically
    converted to a single-item list.

    """
    def __init__(self, *targets):
        self.targets = tuple([t] if isinstance(t, (int, BoundLexicon)) else t
            for t in targets)

    def index(self, match):
        raise NotImplementedError

    def target(self, match):
        return self.targets[self.index(match)]


class MatchTarget(DynamicTarget):
    """A DynamicTarget that calls the predicate with the match object.

    The predicate should return the index of the target to choose.

    """
    def __init__(self, predicate, *targets):
        super().__init__(*targets)
        self.predicate = predicate

    def index(self, match):
        return self.predicate(match)





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module defines the tree structure a text is parsed into.

A tree consists of Context and Token objects.

A Context is a list containing Tokens and other Contexts. A Context is created
when a lexicon becomes active. A Context knows its parent Context and its
lexicon.

A Token represents one parsed piece of text. A Token is created when a rule in
the lexicon matches. A Token knows its parent Context, its position in the text
and the action that was specified in the rule.

The root Context is always one, and it represents the root lexicon. A Context
is always non-empty, except for the root Context, which is empty if the
document did not generate a single token.

The tree structure is very easy to navigate, no special objects or iterators
are necessary for that.

To find a token at a certain position in a context, use find_token() and its
relatives. From every token you can iterate forward() and backward(). Use the
methods like left_siblings() and right_siblings() to traverse the current
context.

See also the documentation for Token and Context.

"""


import sys
import itertools

from livelex import util
from livelex import query


class NodeMixin:
    """Methods that are shared by Token and Context."""
    __slots__ = ()

    is_token = False
    is_context = False

    def parent_index(self):
        """Return our index in the parent.

        This is recommended above using parent.index(self), because this method
        finds our index using a binary search on position, while the latter
        is a linear search, which is certainly slower with a large number of
        children.

        """
        p = self.parent
        pos = self.pos
        lo = 0
        hi = len(p)
        while lo < hi:
            mid = (lo + hi) // 2
            n = p[mid]
            if n.pos < pos:
                lo = mid + 1
            elif n is self:
                return mid
            else:
                hi = mid
        return lo

    def dump(self, depth=0):
        """Prints a nice graphical representation, for debugging purposes."""
        prefix = (" ╰╴" if self.is_last() else " ├╴") if depth else ""
        node = self
        for i in range(depth - 1):
            node = node.parent
            prefix = ("   " if node.is_last() else " │ ") + prefix
        print(prefix + repr(self))

    def root(self):
        """Return the root node."""
        root = self
        for root in self.ancestors():
            pass
        return root

    def is_root(self):
        """Return True if this Node has no parent node."""
        return self.parent is None

    def is_last(self):
        """Return True if this Node is the last child of its parent.

        Fails if called on the root element.

        """
        return self.parent[-1] is self

    def is_first(self):
        """Return True if this Node is the first child of its parent.

        Fails if called on the root element.

        """
        return self.parent[0] is self

    def ancestors(self, upto=None):
        """Climb the tree up over the parents.

        If upto is given and it is one of the ancestors, stop after yielding
        that ancestor. Otherwise iteration stops at the root node.

        """
        node = self.parent
        if upto and upto.parent is not None:
            p = upto.parent
            while node is not None and node is not p:
                yield node
                node = node.parent
        else:
            while node is not None:
                yield node
                node = node.parent

    def ancestors_with_index(self, upto=None):
        """Yield the ancestors(upto), and the index of each node in the parent."""
        n = self
        for p in self.ancestors(upto):
            yield p, n.parent_index()
            n = p

    def common_ancestor(self, other):
        """Return the common ancestor with the Context or Token."""
        ancestors = [self]
        ancestors.extend(self.ancestors())
        if other in ancestors:
            return other
        for n in other.ancestors():
            if n in ancestors:
                return n

    def left_sibling(self):
        """Return the left sibling of this node, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[0] is not self:
            i = self.parent_index()
            return self.parent[i-1]

    def right_sibling(self):
        """Return the right sibling of this node, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[-1] is not self:
            i = self.parent_index()
            return self.parent[i+1]

    def left_siblings(self):
        """Yield the left siblings of this node in reverse order, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[0] is not self:
            i = self.parent_index()
            yield from self.parent[i-1::-1]

    def right_siblings(self):
        """Yield the right siblings of this node, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[-1] is not self:
            i = self.parent_index()
            yield from self.parent[i+1:]

    def next_token(self):
        """Return the following Token, if any."""
        for t in self.forward():
            return t

    def previous_token(self):
        """Return the preceding Token, if any."""
        for t in self.backward():
            return t

    def forward(self, upto=None):
        """Yield all Tokens in forward direction.

        Descends into child Contexts, and ascends into parent Contexts.
        If upto is given, does not ascend above that context.

        """
        for parent, index in self.ancestors_with_index(upto):
            yield from tokens(parent[index+1:])

    def backward(self, upto=None):
        """Yield all Tokens in backward direction.

        Descends into child Contexts, and ascends into parent Contexts.
        If upto is given, does not ascend above that context.

        """
        for parent, index in self.ancestors_with_index(upto):
            if index:
                yield from tokens_bw(parent[index-1::-1])


class Token(NodeMixin):
    """A Token instance represents a lexed piece of text.

    A token has the following attributes:

    `parent`: the Context node to which the token was added
    `pos`:    the position of the token in the original text
    `end`:    the end position of the token in the original text
    `text`:   the text of the token
    `action`: the action specified by the lexicon rule that created the token

    When a pattern rule in a lexicon matches the text, a Token is created.
    When that rule would create more than one Token from a single regular
    expression match, _GroupToken objects are created instead, carrying the
    tuple of all instances in the `group` attribute. The `group` attribute is
    readonly None for normal tokens.

    GroupTokens are thus always adjacent in the same context. If you want to
    retokenize text starting at some position, be sure you are at the start of
    a grouped token, e.g.::

        t = ctx.find_token(45)
        if t.group:
            t = t.group[0]
        pos = t.pos

    (A _GroupToken is just a normal Token otherwise, the reason a subclass was
    created is that the group attribute is unused in by far the most tokens, so
    it does not use any memory. You never need to reference the _GroupToken
    class; just test the group attribute if you want to know if a token belongs
    to a group that originated from a single match.)

    When iterating over the children of a Context (which may be Context or
    Token instances), you can use the `is_token` attribute to determine whether
    the node child is a token, which is easier than to call `isinstance(t,
    Token)` each time.

    From a token, you can iterate `forward()` or `backward()` to find adjacent
    tokens. If you only want to stay in the current context, use the various
    sibling methods, such as `right_sibling()`.

    By traversing the `ancestors()` of a token or context, you can find which
    lexicons created the tokens.

    You can compare a Token instance with a string. Instead of::

        if token.text == "bla":
            do_something()

    you can do::

        if token == "bla":
            do_something()

    You can call `len()` on a token, which returns the length of the token's
    text attribute, and you can use the string format method to embed the
    token's text in another string:

        s = "blabla {}".format(token)

    A token always has a parent, and that parent is always a Context instance.

    """

    __slots__ = "parent", "pos", "text", "action"

    is_token = True
    group = None

    def __init__(self, parent, pos, text, action):
        self.parent = parent
        self.pos = pos
        self.text = text
        self.action = action

    def copy(self):
        """Return a shallow copy.

        The parent still points to the parent of the original.

        """
        return type(self)(self.parent, self.pos, self.text, self.action)

    def equals(self, other):
        """Return True if other has same pos, text and action and state."""
        return (self.pos == other.pos
                and self.text == other.text
                and self.action == other.action
                and self.state_matches(other))

    def __repr__(self):
        text = util.abbreviate_repr(self.text[:31])
        return "<Token {} at {}:{} ({})>".format(text, self.pos, self.end, self.action)

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.text
        return super().__eq__(other)

    def __ne__(self, other):
        if isinstance(other, str):
            return other != self.text
        return super().__ne__(other)

    def __format__(self, formatstr):
        return self.text.__format__(formatstr)

    def __len__(self):
        return len(self.text)

    @property
    def end(self):
        return self.pos + len(self.text)

    def forward_including(self, upto=None):
        """Yield all tokens in forward direction, including self."""
        yield self
        yield from self.forward(upto)

    def backward_including(self, upto=None):
        """Yield all tokens in backward direction, including self."""
        yield self
        yield from self.backward(upto)

    def cut(self):
        """Remove this token and all tokens to the right from the tree."""
        for parent, index in self.ancestors_with_index():
            del parent[index+1:]
        del self.parent[-1] # including ourselves

    def cut_left(self):
        """Remove all tokens to the left from the tree."""
        for parent, index in self.ancestors_with_index():
            del parent[:index]

    def split(self):
        """Split off a new tree, starting with this token.

        The new tree has the same ancestor structure as the current. This token
        and all tokens to the right are moved to the new tree and removed from
        the current one. The new tree's root element is returned.

        """
        parent = self.parent
        node = firstchild = self
        for p, i in self.ancestors_with_index():
            copy = Context(p.lexicon, None)
            copy.append(firstchild)
            if node is not p[-1]:
                s = slice(i + 1, None)
                for n in p[s]:
                    n.parent = copy
                copy.extend(p[s])
                del p[s]
            firstchild.parent = copy
            firstchild = copy
            node = p
        del parent[-1]
        # remove empty context in the old tree
        while not parent and parent.parent:
            parent = parent.parent
            del parent[-1]
        return copy

    def join(self, context):
        """Add ourselves and all tokens to the right to the context.

        This method assumes that the context has the same parent depth
        as our own, and only makes sense if those parents also have the same
        lexicon, i.e. the our state matches the target context (and that
        the pos attribute of the tokens is adjusted).

        The nodes are not removed from their former parents, just the parent
        attribute is changed.

        """
        context.append(self)
        node = self
        c = context
        for p, i in self.ancestors_with_index():
            if node is not p[-1]:
                siblings = p[i+1:]
                for n in siblings:
                    n.parent = c
                c.extend(siblings)
            node = p
            c = c.parent
        self.parent = context

    def state_matches(self, other):
        """Return True if the other Token has the same lexicons in the ancestors."""
        if other is self:
            return True
        for c1, c2 in zip(self.ancestors(), other.ancestors()):
            if c1 is c2:
                return True
            elif c1.lexicon != c2.lexicon:
                return False
        return c1.parent is None and c2.parent is None

    def common_ancestor_with_trail(self, other):
        """Return (context, trail_self, trail_other).

        The context is the common ancestor such as returned by common_ancestor,
        if any. trail_self is a tuple of indices from the common ancestor upto
        self, and trail_other is a tuple of indices from the same ancestor upto
        the other Token.

        If there is no common ancestor, all three are None. But normally,
        all nodes share the root context, so that will normally be the upmost
        common ancestor.

        """
        if other is self:
            i = self.parent_index()
            return self.parent, (i,), (i,)
        s_ancestors, s_indices = zip(*self.ancestors_with_index())
        o_indices = []
        for n, i in other.ancestors_with_index():
            o_indices.append(i)
            try:
                s_i = s_ancestors.index(n)
            except ValueError:
                continue
            return n, s_indices[s_i::-1], o_indices[::-1]
        return None, None, None

    def target(self):
        """Return the first context directly to the right of this Token.

        The context should be the right sibling of the token, or of any of its
        ancestors. If the token is part of a group, the context is found
        immediately next to the last member of the group. The found context may
        also be a child of the grand-parents of this token, in case the target
        popped contexts first.

        In all cases, the returned context is the one started by the target
        in the lexicon rule that created this token.

        """
        node = self
        if node.group:
            node = node.group[-1]
        while node.parent:
            r = node.right_sibling()
            if r:
                if r.is_context:
                    return r
                return
            node = node.parent


class _GroupToken(Token):
    """A Token class that allows setting the `group` attribute."""
    __slots__ = "group"


class Context(list, NodeMixin):
    """A Context represents a list of tokens and contexts.

    The lexicon that created the tokens is in the `lexicon` attribute.

    If a pattern rule jumps to another lexicon, a sub-Context is created and
    tokens are added there. If that lexicon pops back to the current one, new
    tokens can appear after the sub-context. (So the token that caused the jump
    to the sub-context normally preceeds the context it created.)

    A context has a `parent` attribute, which can point to an enclosing
    context. The root context has `parent` None.

    When iterating over the children of a Context (which may be Context or
    Token instances), you can use the `is_context` attribute to determine
    whether the node child is a context, which is easier than to call
    `isinstance(node, Context)` each time.

    You can quickly find tokens in a context::

        if "bla" in context:
            # etc

    And if you want to know which token is on a certain position in the text,
    use e.g.::

        context.find_token(45)

    which, using a bisection algorithm, quickly returns the token, which
    might be in any sub-context of the current context.

    """
    __slots__ = "lexicon", "parent"

    is_context = True

    def __new__(cls, lexicon, parent):
        return list.__new__(cls)

    def __init__(self, lexicon, parent):
        self.lexicon = lexicon
        self.parent = parent

    def copy(self):
        """Return a copy of this Context node, with copies of all the children.

        The parent of the copy is None.

        """
        copy = type(self)(self.lexicon, None)
        for n in self:
            c = n.copy()
            c.parent = copy
            copy.append(c)
        return copy

    def __repr__(self):
        pos, end = self.pos, self.end
        if pos is None:
            pos = end = "?"
        name = self.lexicon and self.lexicon.name()
        return "<Context {} at {}-{} ({} children)>".format(
            name, pos, end, len(self))

    def dump(self, depth=0):
        """Prints a nice graphical representation, for debugging purposes."""
        super().dump(depth)
        for n in self:
            n.dump(depth + 1)

    @property
    def pos(self):
        """Return the position or our first token. Returns None if empty."""
        for t in self.tokens():
            return t.pos

    @property
    def end(self):
        """Return the end position or our last token. Returns None if empty."""
        for t in self.tokens_bw():
            return t.end

    def tokens(self):
        """Yield all Tokens, descending into nested Contexts."""
        yield from tokens(self)

    def tokens_bw(self):
        """Yield all Tokens, descending into nested Contexts, in backward direction."""
        yield from tokens_bw(self[::-1])

    def first_token(self):
        """Return our first Token."""
        for t in self.tokens():
            return t

    def last_token(self):
        """Return our last token."""
        for t in self.tokens_bw():
            return t

    def find_token(self, pos):
        """Return the Token at or to the right of position."""
        i = 0
        hi = len(self)
        while i < hi:
            mid = (i + hi) // 2
            n = self[mid]
            if n.is_context:
                n = n.last_token()
            if n.end <= pos:
                i = mid + 1
            else:
                hi = mid
        if i < len(self):
            if self[i].is_context:
                return self[i].find_token(pos)
            return self[i]
        return self.last_token()

    def find_token_left(self, pos):
        """Return the Token at or to the left of position."""
        i = 0
        hi = len(self)
        while i < hi:
            mid = (i + hi) // 2
            n = self[mid]
            if n.is_context:
                n = n.first_token()
            if n.pos < pos:
                i = mid + 1
            else:
                hi = mid
        if i > 0:
            i -= 1
            if self[i].is_context:
                return self[i].find_token_left(pos)
            return self[i]
        return self.first_token()

    def find_token_after(self, pos):
        """Return the first token completely right from pos.

        Returns None if there is no token right from pos.

        """
        i = 0
        hi = len(self)
        while i < hi:
            mid = (i + hi) // 2
            n = self[mid]
            if n.is_context:
                n = n.last_token()
            if n.pos < pos:
                i = mid + 1
            else:
                hi = mid
        if i < len(self):
            if self[i].is_context:
                return self[i].find_token_after(pos)
            return self[i]

    def find_token_before(self, pos):
        """Return the last token completely left from pos.

        Returns None if there is no token left from pos.

        """
        i = 0
        hi = len(self)
        while i < hi:
            mid = (i + hi) // 2
            n = self[mid]
            if n.is_context:
                n = n.first_token()
            if pos < n.end:
                hi = mid
            else:
                i = mid + 1
        if i > 0:
            i -= 1
            if self[i].is_context:
                return self[i].find_token_before(pos)
            return self[i]

    def tokens_range(self, start, end=None):
        """Yield all tokens in this text range. Use from the root Context."""
        if not self:
            return  # empty
        start_token = self.find_token(start) if start else self.first_token()
        if end is None or end >= self.end:
            yield from start_token.forward_including()
            return
        if start_token.end >= end:
            yield start_token
            return
        end_token = self.find_token_left(end)
        if end_token is start_token:
            yield start_token
            return
        if end_token.pos > start_token.pos:
            # all this is just to avoid a pos comparison on *each* token
            # which is faster for large ranges of tokens
            context, trail_start, trail_end = start_token.common_ancestor_with_trail(end_token)
            # yield start token and then climb down the tree
            yield start_token
            p = start_token.parent
            for i in trail_start[:0:-1]:
                yield from tokens(p[i+1:])
                p = p.parent
            # yield intermediate tokens
            yield from tokens(context[trail_start[0]+1:trail_end[0]])
            # then climb up the tree for the end token
            node = context[trail_end[0]]
            for i in trail_end[1:]:
                yield from tokens(node[:i])
                node = node[i]
            yield node  # == end_token

    def source(self):
        """Return the first Token, if any, when going to the left from this context.

        The returned token is the one that created us, that this context the
        target is for. If the token is member of a group, the first group member
        is returned.

        """
        node = self
        for parent in node.ancestors():
            n = node.left_sibling()
            if n:
                while n:
                    if n.is_token:
                        if n.group:
                            n = n.group[0]
                        return n
                    n = n[-1]
                return
            node = parent

    @property
    def query(self):
        """Query this node in different ways; see the query module."""
        def gen():
            yield self
        return query.Query(gen)


def tokens(nodes):
    """Helper to yield tokens from the iterable of nodes."""
    for n in nodes:
        if n.is_token:
            yield n
        else:
            yield from n.tokens()


def tokens_bw(nodes):
    """Helper to yield tokens from the iterable in backward direction.

    Make sure nodes is already in backward direction.

    """
    for n in nodes:
        if n.is_token:
            yield n
        else:
            yield from n.tokens_bw()





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Helper objects to construct regular expressions.

"""


class Pattern:
    """Base class for objects that build a regular expression."""
    def build(self):
        raise NotImplementedError


class Words(Pattern):
    """Creates a regular expression from a list of words."""
    def __init__(self, words, prefix="", suffix=""):
        self.words = words
        self.prefix = prefix
        self.suffix = suffix

    def build(self):
        from . import regex
        expr = regex.words2regexp(self.words)
        if self.prefix or self.suffix:
            return self.prefix + '(?:' + expr + ')' + self.suffix
        return expr


class Char(Pattern):
    """Create a regular expression matching one of the characters in the string.

    If positive is False, the expression is negated, i.e. to match one character
    if it is not in the string.

    """
    def __init__(self, chars, positive=True):
        self.chars = chars
        self.positive = positive

    def build(self):
        from . import regex
        negate = "" if self.positive else "^"
        return '[' + negate + regex.make_charclass(set(self.chars)) + ']'





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
An experimental query module.


Using this module you can query the token tree to find tokens and contexts,
based on lexicons and/or actions and text contents. You can chain calls
in an XPath-like fashion.

This module supplements the various find_xxx methods of every Context object.
A query starts at the `query` property of a Context object, and yields all the
nodes of that Context.

You can navigate using `children`, `all`, `[n]`, `[n:n]`, (`[n:n:n]`),
`next`, `previous`, `right`, `left`, and `parent`. Use `uniq` to remove double
occurrences of nodes, which can e.g. happen when navigating to the parent of
all nodes.

You can narrow down the search using `tokens`, `contexts`, `('text')` or
`(lexicon)`, `has_not`, `[n]`, `[n:n]`, `[action]`, `[action, action]`,
`startingwith()`, `endingwith()`, `containing()`, `matching()`, `uniq`,
`in_action()`,  and the corresponding `not_` counterparts, and `__getitem__`
and `is_not()`.



Examples:

Find all tokens that are the first child of a Context with bla lexicon:

    root.query.all(MyLang.bla)[0]


Find (in Xml) all attributes with name 'name' that are in a <bla> tag:

    root.query.all[Name.Tag]("bla").next('name')


Find all tags containing "hi" in their text nodes:

    root.query.all[Name.Tag].next.next[Text].containing('hi')


Find all comments that have TODO in it:

    root.query.all[Comment].containing('TODO')


A query is a generator, you can iterate over the results. For debugging
purposes, there are also the list(), pick(), count() and dump() methods.

    for attrs in q.all[Name.Tag]('origin').right:
        for atr in attrs.query[Name.Attribute]:
            print(atr)


Summary of the query methods:

Endpoint methods, mainly for debugging:

    count()
        Just prints the number of nodes in the result set

    dump()
        dump()s the full result nodes to stdout

    list()
        aggregate the results in a list

    pick(default=None)
        just pick the first result, or a default if no results


Navigating nodes:

    The query itself yields al children of the Context it was started from.
    But using the following methods you can find your way through a tree
    structure. Every method returns a new Query object, having the previous
    one as source of nodes. Most methods are implemented as properties, so
    you don't have to write parentheses.

    all
        yield all descendant nodes, depth-first, in order. First it yields the
        context, then its children.

    children
        yield all the direct children of the current nodes

    parent
        yield the parent of all current nodes. This can yield double
        occurrences of nodes in the list. (Use uniq to fix that.)

    next, previous
        yield the next or previous Token from the current node, if any

    right, left
        yield the right or left sibling of every current node, if any

    [int], __getitem__(int)
        yield the nth child (if available) of each Context node
        (supports negative indices)

    [slice]
        yield from the specified slice of each Context node

    first, last
        yield the first resp. the last child of every Context node.
        Same as [0] or [-1].

    target
        yield the target context for a token, if any. See Token.target().

    source
        yield the source token for a context, if any. See Context.source().


Selecting (filtering) nodes:

    These methods filter out current nodes without adding new nodes
    to the selection.

    tokens
        select only the tokens

    contexts
        select only the contexts

    uniq
        Removes double occurrences of Tokens or Contexts, which can happen
        e.g. when selecting the parent of all nodes

    filter(predicate)
        select nodes for which the predicate function returns a value that
        evaluates to True

    map(function)
        call function on every node and yield its results, which should be
        nodes as well.

    is_not
        inverts the meaning of the following query, e.g. is_not.startingwith()

    The following query methods are inverted by `is_not`:

    in_range(start=0, end=None)
        select only the nodes that fully fit in the text range. If preceded
        by `is_not`, selects the nodes that are outside the specified text
        range.

    (lexicon, [lexicon, ...])
        select the Contexts with that lexicon (or one of the lexicons)

    ("text"), ("text", ["text2", ...])
        select the Tokens with exact that text (or one of the texts)

    startingwith("text")
        select the Tokens that start with the specified text

    endingwith("text")
        select the Tokens that end with the specified text

    containing("text")
        select the Tokens that contain specified text

    matching("regex")
        select the Tokens that match the specified regular epression
        (using re.search, the expression can match anywhere unless you use
        ^ or $ characters).

    [action], [action, action<, action> ...]
        select the Tokens that have one of the specified actions

    in_action(*actions)
        select tokens if their action belongs in the realm of one of the
        specified StandardActions

    remove_ancestors
        remove Context nodes from the current node list that have descendants
        in the list.

    remove_descendants
        remove nodes from the current list if any of their ancestors is also
        in the list.


"""


import functools
import re
import sys

from .lexicon import BoundLexicon


def query(func):
    """Make a method result (generator) into a new Query object."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        return Query(lambda: func(self, *args, **kwargs))
    return wrapper


def pquery(func):
    """Make a method result into a Query object, and the method a property."""
    return property(query(func))


class Query:
    __slots__ = '_gen', '_inv'

    def __init__(self, gen, invert=False):
        self._gen = gen
        self._inv = invert

    def __iter__(self):
        return self._gen()

    @property
    def is_not(self):
        """Invert the next query."""
        return Query(self._gen, not self._inv)

    # end points
    def count(self):
        """Compute the length of the iterable. Don't use this other than for debugging."""
        return sum(1 for _ in self)

    def dump(self):
        """Dump the current selection to the console."""
        for n in self:
            n.dump()

    def list(self):
        """Return the current selection as a list. Only for debugging."""
        return list(self)

    def pick(self, default=None):
        """Pick the first value, or return the default."""
        for n in self:
            return n
        return default

    # navigators
    @pquery
    def children(self):
        """All direct children of the current nodes."""
        for n in self:
            if n.is_context:
                yield from n

    @pquery
    def all(self):
        """All descendants, contexts and their nodes."""
        def innergen(node):
            for n in node:
                yield n
                if n.is_context:
                    yield from innergen(n)
        return innergen(self)

    @pquery
    def parent(self):
        """The parent of every node."""
        for n in self:
            if n.parent:
                yield n.parent

    @pquery
    def first(self):
        """Yield the first node of every context node, same as [0]."""
        for n in self:
            if n and n.is_context:
                yield n[0]

    @pquery
    def last(self):
        """Yield the last node of every context node, same as [-1]."""
        for n in self:
            if n and n.is_context:
                yield n[-1]

    @pquery
    def next(self):
        """Yield the next token, if any."""
        for n in self:
            t = n.text_token()
            if t:
                yield t

    @pquery
    def previous(self):
        """Yield the previous token, if any."""
        for n in self:
            t = n.previous_token()
            if t:
                yield t

    @pquery
    def right(self):
        """Yield the right sibling, if any."""
        for n in self:
            n = n.right_sibling()
            if n:
                yield n

    @pquery
    def left(self):
        """Yield the left sibling, if any."""
        for n in self:
            n = n.left_sibling()
            if n:
                yield n

    @pquery
    def target(self):
        """Yield the target Context for every token, if available. See Token.target()."""
        for t in self:
            if t.is_token:
                target = t.target()
                if target:
                    yield target

    @pquery
    def source(self):
        """Yield the source Token for every context, if available. See Context.source()."""
        for n in self:
            if n.is_context:
                source = n.source()
                if source:
                    yield source

    @query
    def map(self, function):
        """Call the function on every node and yield its results, which should be zero or more nodes as well."""
        for n in self:
            yield from function(n)

    # selectors
    @query
    def filter(self, predicate):
        """Yield nodes for which the predicate returns a value that evaluates to True."""
        for n in self:
            if predicate(n):
                yield n

    @pquery
    def tokens(self):
        """Get only the tokens."""
        for n in self:
            if n.is_token:
                yield n

    @pquery
    def contexts(self):
        """Get only the contexts."""
        for n in self:
            if n.is_context:
                yield n

    @pquery
    def uniq(self):
        """Remove double occurrences. Can happen when you use the parent."""
        seen = set()
        for n in self:
            i = id(n)
            if i not in seen:
                seen.add(i)
                yield n

    @pquery
    def remove_descendants(self):
        """Remove nodes that have ancestors in the current node list."""
        ids = set(map(id, self))
        for n in self:
            if not any(id(p) in ids for p in n.ancestors()):
                yield n

    @pquery
    def remove_ancestors(self):
        """Remove nodes that have descendants in the current node list."""
        ids = set(map(id, self)) & set(map(id, (p for n in self for p in n.ancestors())))
        for n in self:
            if id(n) not in ids:
                yield n

    # invertible selectors
    @query
    def in_range(self, start=0, end=None):
        """Yield a restricted set, tokens and/or contexts must fall in start→end"""
        if end is None:
            end = sys.maxsize
        # don't assume the tokens are in source order
        if self._inv:
            for n in self:
                if n.end <= start or n.pos >= end:
                    yield n
        else:
            for n in self:
                if n.pos >= start and n.end <= end:
                    yield n

    @query
    def __call__(self, *what):
        """Yield token if token has that text, or context if context has that lexicon."""
        if isinstance(what[0], BoundLexicon):
            for n in self:
                if n.is_context and self._inv ^ (n.lexicon in what):
                    yield n
        else:
            for n in self:
                if n.is_token and self._inv ^ (n.text in what):
                    yield n

    @query
    def startingwith(self, text):
        """Yield tokens that start with text."""
        for t in self:
            if t.is_token and self._inv ^ t.text.startswith(text):
                yield t

    @query
    def endingwith(self, text):
        """Yield tokens that end with text."""
        for t in self:
            if t.is_token and self._inv ^ t.text.endswith(text):
                yield t

    @query
    def containing(self, text):
        """Yield tokens that contain the specified text."""
        for t in self:
            if t.is_token and self._inv ^ (text in t.text):
                yield t

    @query
    def __getitem__(self, key):
        """normal slicing, and you can test for one or more actions."""
        # slicing or itemgetting with integers are not invertible selectors
        if isinstance(key, int):
            for n in self:
                if n.is_context:
                    if key < 0:
                        key += len(n)
                    if 0 <= key < len(n):
                        yield n[key]
        elif isinstance(key, slice):
            for n in self:
                if n.is_context:
                    yield from n[key]
        # handle matching actions, this is invertible by is_not
        elif isinstance(key, tuple):
            for t in self:
                if t.is_token and self._inv ^ (t.action in key):
                    yield t
        else:
            for t in self:
                if t.is_token and self._inv ^ (t.action is key):
                    yield t

    @query
    def matching(self, pattern, flags=0):
        """Yield tokens matching the regular expression (using re.search)."""
        for t in self:
            if t.is_token and self._inv ^ bool(re.search(pattern, t.text, flags)):
                yield t

    @query
    def in_action(self, *actions):
        """Yield those tokens whose action is or inherits from one of the given actions."""
        for t in self:
            if t.is_token and self._inv ^ any(t.action in a for a in actions):
                yield t




# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module contains some smart functions to build regular expressions:

* words2regexp() builds an optimized regular expression from a list of words.

The following helper functions are available:

* common_suffix() returns the common suffix of the words, if any.
* make_trie() makes a dict-based radix tree structure from a list of words.
* trie_to_regexp_tuple() converts a trie structure to a regexp tuple.
* build_regexp() converts a tuple to a real regular expression.
* make_charclass() turns a list of characters into a sensible [a-z] expression.

"""


import re


def words2regexp(words):
    """Convert the word list to an optimized regular expression."""
    words, suffix = common_suffix(words)
    root = make_trie(words)
    r = trie_to_regexp_tuple(root)
    if suffix:
        r += (suffix,)
    return build_regexp(r)


def trie_to_regexp_tuple(node, reverse=False):
    """Converts the trie node to a tuple of regular expression parts.

    A part is either a plain string expression or a frozenset instance.
    A frozenset instance denotes a group of alternative expressions, and
    consists of plain string expressions or other tuples. If None is also
    present in the frozenset, the expression is optional.

    """
    if reverse:
        combine = lambda r1, r2: r2 + r1
    else:
        combine = lambda r1, r2: r1 + r2

    if len(node) == 1:
        for k, n in node.items():
            if k:
                return combine((k,), trie_to_regexp_tuple(n, reverse))
            return ()
    else:
        seen = []
        keys = []
        groups = set()

        # group the nodes if they have the same leaf node
        for k, n in node.items():
            if k:
                try:
                    i = seen.index(n)
                except ValueError:
                    i = len(seen)
                    seen.append(n)
                    keys.append([k])
                else:
                    keys[i].append(k)
            else:
                groups.add(None)    # means optional group, may end here

        for keys, node in zip(keys, seen):
            if len(keys) == 1:
                if not any(node):
                    groups.add(keys[0])
                else:
                    groups.add(combine((keys[0],), trie_to_regexp_tuple(node, reverse)))
            else:
                if not reverse:
                    # try to optimize the keys backwards
                    r = trie_to_regexp_tuple(make_trie(keys, True), True)
                    if r == (frozenset(keys),) and not any(node):
                        groups.update(keys)
                        continue
                elif not any(node):
                    groups.update(keys)
                    continue
                else:
                    r = (frozenset(keys),)
                groups.add(combine(r, trie_to_regexp_tuple(node, reverse)))
        return groups.pop() if len(groups) == 1 else (frozenset(groups),)


def build_regexp(r):
    """Convert a tuple to a full regular expression pattern string.

    The tuple is described in the trie_to_regexp_tuple() function doc string.

    """
    def get_items(r):
        """Yield regexp items from tuple r in tuples (item, mincount, maxcount).

        An item is either a string like "aa", or a two-tuple(exprs,
        tuples), where exprs is a set of plain strings, and tuples a set of
        tuples that were inside a frozenset. The mincount and maxcount are
        integers and describe the minimal or maximal required count of
        matches.

        """
        for item in r:
            if isinstance(item, str):
                yield item, 1, 1
            else:
                # item is a frozenset
                mincount = 1
                exprs = set()
                tuples = set()
                for k in item:
                    if isinstance(k, str):
                        exprs.add(k)
                    elif isinstance(k, tuple):
                        tuples.add(k)
                    elif k is None:
                        mincount = 0
                # just one expression and no subgroups?
                if len(exprs) == 1 and not tuples:
                    yield exprs.pop(), mincount, 1
                else:
                    item = (exprs, tuples)
                    # optimize for the case of only one subgroup
                    # remove otherwise empty parent group if possible
                    if not exprs and len(tuples) == 1:
                        # there is only one subexpression in the group
                        r = next(iter(tuples))
                        items = merge_items(r)
                        # if our group has no qualifier, just yield the subgroup
                        if mincount:
                            yield from items
                        # if we are optional, check if the qualifier of the subgroup
                        # can be altered. Possible when mincount <= 1.
                        elif len(items) == 1 and items[0][1] <= 1:
                            item, _, maxcount = items[0]
                            yield item, 0, maxcount
                        else:
                            yield item, mincount, 1
                    else:
                        yield item, mincount, 1

    def merge_items(r):
        """Read items-tuples such as yielded by get_items().

        Returns a list of the same items, merging where possible adjacent
        items that are the same. Every list entry is a three-tuple(item,
        mincount, maxcount); where an item is either a string like "aa", or
        a two-tuple(exprs, tuples).

        """
        items = []
        for item, mincount, maxcount in get_items(r):
            if items and items[-1][0] == item:
                items[-1][1] += mincount
                items[-1][2] += maxcount
            else:
                items.append([item, mincount, maxcount])
        return items

    items = merge_items(r)

    # now really construct the regexp string for each item
    result = []
    for item, mincount, maxcount in items:
        # qualifier to use
        if mincount == 1 and maxcount == 1:
            qualifier = ''
        elif mincount == 0 and maxcount == 1:
            qualifier = "?"
        elif mincount == maxcount:
            qualifier = "{{{0}}}".format(maxcount)
        else:
            qualifier = "{{{0},{1}}}".format(mincount or '', maxcount or '')
        # make the rx
        if isinstance(item, str):
            rx = re.escape(item)
            enclose = len(item) > 1 and qualifier
        else:
            exprs, tuples = item
            # separate single characters from longer strings
            chars, strings = set(), set()
            for k in exprs:
                (chars if len(k) == 1 else strings).add(k)
            group = []
            if chars:
                if len(chars) == 1:
                    group.append(re.escape(next(iter(chars))))
                else:
                    group.append('[' + make_charclass(chars) + ']')
            if strings:
                group.extend(map(re.escape, sorted(strings)))
            if tuples:
                group.extend(map(build_regexp, tuples))
            if chars and not strings and not tuples:
                rx = group[0]
                enclose = False
            else:
                rx = '|'.join(group)
                enclose = len(items) > 1 or qualifier
        if enclose:
            rx = '(?:' + rx + ')'
        result.append(rx + qualifier)
    return ''.join(result)


def make_charclass(chars):
    """Return a string with adjacent characters grouped.

    eg ('a', 'b', 'c', 'd', 'f') is turned into '[a-df]'.
    Special characters are properly escaped.

    """
    buf = []
    for c in sorted(map(ord, chars)):
        if buf and buf[-1][1] == c - 1:
            buf[-1][1] = c
        else:
            buf.append([c, c])
    return ''.join(re.escape(chr(a)) if a == b else
                   re.escape(chr(a) + chr(b)) if a == b - 1 else
                   re.escape(chr(a)) + '-' + re.escape(chr(b))
                   for a, b in buf)


def make_trie(words, reverse=False):
    """Return a dict-based radix trie structure from a list of words.

    If reverse is set to True, the trie is made in backward direction,
    from the end of the words.

    """
    if reverse:
        chars = lambda word: word[::-1]
        add = lambda k1, k2: k2 + k1
    else:
        chars = lambda word: word
        add = lambda k1, k2: k1 + k2

    root = {}
    for w in words:
        d = root
        for c in chars(w):
            d = d.setdefault(c, {})
        d[None] = True  # end

    # merge characters that are the only child with their parents
    def merge(node):
        for key, node in node.items():
            if key:
                while len(node) == 1:
                    k, n = next(iter(node.items()))
                    if k:
                        key = add(key, k)
                        node = n
                    else:
                        break
                else:
                    node = dict(merge(node))
            yield key, node

    return dict(merge(root))


def common_suffix(words):
    """Return (words, suffix), where suffix is the common suffix.

    If there is no common suffix, words is unchanged, and suffix is an
    empty string. If there is a common suffix, that is chopped of the returned
    words.

    """
    suffix = []
    for s in map(set, zip(*map(reversed, words))):
        if len(s) != 1:
            break
        suffix.append(s.pop())
    suffix = ''.join(reversed(suffix))
    if suffix:
        i = -len(suffix)
        words = [word[:i] for word in words]
    return words, suffix


def to_string(expr):
    """If the regular expression can be converted to a simple string, return it.

    Otherwise, None is returned. The returned string can be used with
    "".find(), which would be faster than using re.search().

    """
    if set(re.sub(r'\\.', '', expr)) & set("^$|.()[]{}+*?"):
        return  # there are unescaped special characters like (, [, ? etc.
    # handle all escapes, there may be fails, in that case we can't use the expr as string
    pat = (r'\\(?:'
        r'x([0-9a-fA-F]{2})'    # 1 hex
        r'|([afnrtv])'         # 2 normal escaped character like \n
        r'|(\d{1,3})'           # 3 octal , or fail if back ref
        r'|u([0-9a-fA-F]{4})'   # 4 \uxxxx
        r'|U([0-9a-fA-F]{8})'   # 5 \Uxxxxxxxx
        r'|([\^\$\|\.\(\)\[\]\{\}\+\*\?\\])'  # special re char that was escaped
        r'|)')
    def replace_escapes(m):
        if m.group(1):
            return chr(int(m.group(1), 16))
        elif m.group(2):
            return chr({'a':7, 'f':12, 'n':10, 'r':13, 't':9, 'v':11}[m.group(2)])
        elif m.group(3):
            if 0 < int(m.group(3)) <= 99 and not m.group(3).startswith("0"):
                raise ValueError
            return chr(int(m.group(3), 8))  # can also raise ValueError
        elif m.group(4):
            return chr(int(m.group(4), 16))
        elif m.group(5):
            return chr(int(m.group(5), 16))
        elif m.group(6):
            return m.group(6)
        raise ValueError
    try:
        s = re.sub(pat, replace_escapes, expr)
    except ValueError:
        return
    assert re.fullmatch(expr, s)
    return s





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The livelex Python module.

"""

from . import pattern, action, target
from . import lexicon as lexicon_, treebuilder, document, treedocument
from .document import Cursor
from .language import Language
from .pkginfo import version, version_string


# these can be used in rules where a pattern is expected
default_action = object()   # denotes a default action for unmatched text
default_target = object()   # denotes a default target when no text matches


# used to suppress a token
skip = action.SkipAction()

# predefined standard actions
Whitespace = action.StandardAction("Whitespace")
Text = action.StandardAction("Text")

Escape = action.StandardAction("Escape")
Keyword = action.StandardAction("Keyword")
Name = action.StandardAction("Name")
Literal = action.StandardAction("Literal")
Delimiter = action.StandardAction("Delimiter")
Comment = action.StandardAction("Comment")
Error = action.StandardAction("Error")

Verbatim= Literal.Verbatim
String = Literal.String
Number = Literal.Number
Boolean = Literal.Boolean
Char = Literal.Char
Operator = Delimiter.Operator
Builtin = Name.Builtin
Function = Name.Function
Variable = Name.Variable


class Document(treedocument.TreeDocumentMixin, document.Document):
    """A Document that automatically keeps its contents tokenized."""
    def __init__(self, root_lexicon=None, text=""):
        document.Document.__init__(self, text)
        builder = treebuilder.BackgroundTreeBuilder(root_lexicon)
        treedocument.TreeDocumentMixin.__init__(self, builder)
        if text:
            with builder.change() as c:
                c.change_contents(text)


def words(words):
    """Return a Pattern object that builds a regular expression from a list of words.

    To be used as a pattern in a rule.

    """
    return pattern.Words(words)


def char(chars, positive=True):
    """Return a Pattern matching one of the characters in the specified string.

    If positive is False, the set of characters is complemented, i.e. the
    Pattern matches any single character that is not in the specified string.

    """
    return pattern.Char(chars, positive)


def bygroup(*actions):
    """Return a SubgroupAction that yields tokens for each subgroup in a regular expression.

    To be used as an action in a rule.

    """
    return action.SubgroupAction(*actions)


def bymatch(predicate, *actions):
    """Return a MatchAction that chooses the action based on the match object.

    To be used as an action in a rule.

    The predicate is run with the match object as argument and should return
    the (integer) index of the desired action. True and False are also valid
    return values, they count as 1 and 0, respectively.

    This might look cumbersome (it might look easier to just return the action
    from the function), but this way we know the possible actions beforehand,
    and we could translate the actions via a mapping and still keep everything
    working.

    """
    return action.MatchAction(predicate, *actions)


def bytext(predicate, *actions):
    """Return a TextAction that chooses the action based on the text.

    To be used as an action in a rule.

    The predicate is run with the match object as argument and should return
    the index of the desired action. See also bymatch().

    """
    return action.TextAction(predicate, *actions)


def tomatch(predicate, *targets):
    """Return a MatchTarget that chooses the target based on the match object.

    To be used as a target in a rule.

    The predicate is run with the match object as argument and should return
    the (integer) index of the desired target. True and False are also valid
    return values, they count as 1 and 0, respectively.

    Each target can be a list or tuple of targets, just like in normal rules,
    where the third and more items form a list of targets. A target can also be
    a single integer or a lexicon. Use () or 0 for a target that does nothing.

    """
    return target.MatchTarget(predicate, *targets)


def lexicon(rules_func=None, **kwargs):
    """Lexicon factory decorator.

    Use this decorator to make a function in a Language class definition a
    Lexicon object. The Lexicon object is actually a descriptor, and when
    calling it via the Language class attribute, a BoundLexicon is created,
    cached and returned.

    You can specify keyword arguments, that will be passed on to the
    BoundLexicon object as soon as it is created.

    The following keyword arguments are supported:

    re_flags: The flags that are passed to the regular expression compiler

    The code body of the function should return (yield) the rules of the
    lexicon, and is run with the Language class as first argument, as soon as
    the lexicon is used for the first time.

    You can also call the BoundLexicon object just as an ordinary classmethod,
    to get the rules, e.g. for inclusion in a different lexicon.

    """
    if rules_func and not kwargs:
        return lexicon_.Lexicon(rules_func)
    def lexicon(rules_func):
        return lexicon_.Lexicon(rules_func, **kwargs)
    return lexicon


def root(root_lexicon, text):
    """Return the root context of the tree structure of all tokens from text."""
    return treebuilder.TreeBuilder(root_lexicon).tree(text)


def tokens(root_lexicon, text):
    """Convenience function that yields all the tokens from the text."""
    return root(root_lexicon, text).tokens()





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Document and Cursor form the basis of handling of documents in the livelex
package.

A Document contains a text string that is mutable via item and slice methods.

If you make modifications while inside a context (using the Python context
manager protocol), the modifications are only applied when the context
exits for the last time.

You can use a Cursor to keep track of positions in a document. The position
(and selection) of a Cursor is adjusted when the text in the document is
changed.

For tokenized documents, livelex inherits from this base class.

"""


import itertools
import re
import weakref


from . import util


class AbstractDocument:
    """Base class for a Document.

    To make a Document work, you should at least implement:

        text()
        _update_contents()

    The method text() should simply return the entire text string.

    The method _update_contents() should read the (start, end, text) tuples
    from the list in self._changes, which is already sorted. These changes
    will never overlap. All start/end positions refer to the original state
    of the text.

    For efficiency reasons, you might want to reimplement:

        set_text()
        __len__()
        _get_contents() (called by __getitem__)

    """
    def __init__(self):
        self._cursors = weakref.WeakSet()
        self._edit_context = 0
        self._changes = []

    def text(self):
        """Should return the text."""
        raise NotImplementedError

    def set_text(self, text):
        """Set the text."""
        assert self._edit_context == 0, "can't use set_text() in edit context."
        self[:] = text

    def __repr__(self):
        text = util.abbreviate_repr(self[:31])
        return "<{} {}>".format(type(self).__name__, text)

    def __str__(self):
        return self.text()

    def __format__(self, formatstr):
        return format(self.text(), formatstr)

    def __len__(self):
        return len(self.text())

    def __enter__(self):
        """Start the context for modifying the document."""
        self._edit_context += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context for modifying."""
        if exc_type is not None:
            # cancel all edits when an exception occurred
            self._edit_context = 0
            self._changes.clear()
        elif self._edit_context > 1:
            self._edit_context -= 1
        else:
            self._edit_context = 0
            self._apply_changes()

    def __setitem__(self, key, text):
        start, end = self._parse_key(key)
        if ((text or start != end) and
            (end - start != len(text) or self[start:end] != text)):
            self._changes.append((start, end, text))
            if not self._edit_context:
                self._apply_changes()

    def __delitem__(self, key):
        self[key] = ""

    def __getitem__(self, key):
        start, end = self._parse_key(key)
        if start == end:
            return ""
        elif start == 0 and end == len(self):
            return self.text()
        return self._get_contents(start, end)

    def _parse_key(self, key):
        """Get start and end values from key. Called by __[gs]etitem__."""
        total = len(self)
        if isinstance(key, slice):
            start = key.start or 0
            end = key.stop
        elif isinstance(key, Cursor):
            start = key.start
            end = key.end
        else:
            # single integer
            if key < 0:
                key += total
            if 0 <= key < total:
                return key, key + 1
            raise IndexError("index out of range")
        if start is None or start < -total:
            start = 0
        elif start < 0:
            start += total
        elif start > total:
            start = total
        if end is None or end > total:
            end = total
        elif end < -total:
            end = 0
        elif end < 0:
            end += total
        if end < start:
            end = start
        return start, end

    def _get_contents(self, start, end):
        """Return the selected range of the text.

        Called by __getitem__(), only if a fragment was requested.

        """
        return self.text()[start:end]

    def _apply_changes(self):
        """Apply the changes and update the positions of the cursors."""
        if self._changes:
            self._changes.sort()
            # check for overlaps, find the region
            head = old = self._changes[0][0]
            added = 0
            for start, end, text in self._changes:
                if start < old:
                    raise RuntimeError("overlapping changes: {}".format(self._changes))
                added += start - old + len(text)
                old = end
            self._update_cursors()
            self._update_contents()
            self._changes.clear()
            self.contents_changed(head, end - head, added)

    def _update_cursors(self):
        """Update the positions of the cursors."""
        i = 0
        cursors = sorted(self._cursors, key = lambda c: c.start)
        for start, end, text in self._changes:
            for c in cursors[i:]:
                ahead = c.start > start
                if ahead:
                    if end >= c.start:
                        c.start = start
                    else:
                        c.start += start + len(text) - end
                if c.end is not None and c.end >= start:
                    if end >= c.end:
                        c.end = start + len(text)
                    else:
                        c.end += start + len(text) - end
                elif not ahead:
                    i += 1  # don't consider this cursor any more

    def _update_contents(self):
        """Should apply the changes (in self._changes) to the text."""
        raise NotImplementedError

    def insert(self, pos, text):
        """Insert text at pos."""
        if text:
            self[pos:pos] = text

    def replace(self, old, new, start=0, end=None, count=0):
        """Replace occurrences of old with new in region start->end.

        If count > 0, specifies the maximum number of occurrences to be
        replaced.

        """
        if old == new:
            return
        text = self[start:end]
        length = len(old)
        with self:
            pos = text.find(old)
            while pos >= 0:
                self[start+pos:start+pos+length] = new
                pos = text.find(old, pos + 1)
                count -= 1
                if count == 0:
                    break

    def re_sub(self, pattern, replacement, start=0, end=None, count=0, flags=0):
        """Replace regular expression matches of pattern with replacement.

        Backreferences are allowed. The region can be set with start and end.
        If count > 0, specifies the maximum number of occurrences to be
        replaced.

        The replacement argument can also be a funtion, which is then called
        with the match object and should return the replacement string.

        """
        text = self[start:end]
        if not callable(replacement):
            replacement = (lambda repl: lambda m: m.expand(repl))(replacement)
        with self:
            for i, m in enumerate(re.finditer(pattern, text, flags), 1):
                self[m.start():m.end()] = replacement(m)
                if i == count:
                    break

    def trim(self, start=0, end=None):
        """Remove trialing whitespace in the specified region."""
        self.re_sub(r'\s+$', '', start, end, flags=re.MULTILINE)

    def translate(self, mapping, start=0, end=None, count=0, whole_words=False):
        """Replace every occurrence of a key in mapping with its value.

        If whole_words is True, only match the keys at word boundaries.

        """
        from . import regex
        expr = regex.words2regexp(mapping.keys())
        if whole_words:
            expr = r"\b({})\b".format(expr)
        repl = lambda m: mapping[m.group()]
        self.re_sub(expr, repl, start, end, count)

    def contents_changed(self, position, removed, added):
        """Called by _apply(). The default implementation does nothing."""
        pass


class Document(AbstractDocument):
    """A basic Document with undo and modified status.

    This Document implements AbstractDocument by holding the text in a hidden
    _text attribute. It also adds support for undo/redo and has a modified()
    state.

    A Document is like a mutable string. E.g.:

    .. code-block:: python

        d = Document('some string')
        with d:
            d[5:5] = 'different '
        d.text()  --> 'some different string'

    You can also make modifications outside the a context, they will then
    be applied immediately, which is slower.

    You can enter a context multiple times, and changes will be applied when the
    last exits.

    """
    undo_redo_enabled = True

    def __init__(self, text=""):
        super().__init__()
        self._text = text
        self._modified = False
        self._undo_stack = []
        self._redo_stack = []
        self._in_undo = None

    def modified(self):
        """Return whether the text was modified."""
        return self._modified

    def set_modified(self, modified):
        """Sets whether the text is modified, happens automatically normally."""
        self._modified = modified
        if not modified and not self._in_undo:
            self._set_all_undo_redo_modified()

    def text(self):
        """Return all text."""
        return self._text

    def _update_contents(self):
        """Apply the changes to the text."""
        result = []
        head = tail = self._changes[0][0]
        for start, end, text in self._changes:
            if start > tail:
                result.append(self._text[tail:start])
            if text:
                result.append(text)
            tail = end
        text = "".join(result)
        if self.undo_redo_enabled:
            # store start, end, and text needed to undo this change
            self._handle_undo(head, head + len(text), self._text[head:tail])
        self._text = self._text[:head] + text + self._text[tail:]
        if not self._in_undo:
            self.set_modified(True) # othw this is handled by undo/redo

    def _handle_undo(self, start, end, text):
        """Store start, end, and text needed to reconstruct the previous state."""
        if self._in_undo == "undo":
            self._redo_stack.append([start, end, text, self.modified()])
        else:
            self._undo_stack.append([start, end, text, self.modified()])
            if self._in_undo != "redo":
                self._redo_stack.clear()

    def _set_all_undo_redo_modified(self):
        """Called on set_modified(False). Set all undo/redo state to modified."""
        for undo in itertools.chain(self._undo_stack, self._redo_stack):
            undo[3] = True

    def undo(self):
        """Undo the last modification."""
        assert self._edit_context == 0, "can't undo while in edit context"
        if self._undo_stack:
            self._in_undo = "undo"
            start, end, text, modified = self._undo_stack.pop()
            self[start:end] = text
            self.set_modified(modified)
            self._in_undo = None

    def redo(self):
        """Redo the last undone modification."""
        assert self._edit_context == 0, "can't redo while in edit context"
        if self._redo_stack:
            self._in_undo = "redo"
            start, end, text, modified = self._redo_stack.pop()
            self[start:end] = text
            self.set_modified(modified)
            self._in_undo = None

    def clear_undo_redo(self):
        """Clear the undo/redo stack."""
        self._undo_stack.clear()
        self._redo_stack.clear()

    def can_undo(self):
        """Return True whether undo is possible."""
        return bool(self._undo_stack)

    def can_redo(self):
        """Return True whether redo is possible."""
        return bool(self._redo_stack)


class Cursor:
    """Describes a certain range (selection) in a Document.

    You may change the start and end attributes yourself. Both must be an
    integer, end may also be None, denoting the end of the document.

    As long as you keep a reference to the Cursor, its positions are updated
    when the document changes. When text is inserted at the start position, the
    position remains the same. But when text is inserted at the end of a
    cursor, the end position moves along with the new text. E.g.:

    .. code-block:: python

        d = Document('hi there, folks!')
        c = Cursor(d, 8, 8)
        with d:
            d[8:8] = 'new text'
        c.start, c.end --> (8, 16)

    You can also use a Cursor as key while editing a document:

    .. code-block:: python

        c = Cursor(d, 8, 8)
        with d:
            d[c] = 'new text'

    You cannot alter the document via the Cursor.

    """
    __slots__ = "_document", "end", "start", "__weakref__"

    def __init__(self, document, start=0, end=-1):
        """Init with document. Start default to 0 and end defaults to start."""
        self._document = document
        self.start = start
        self.end = end if end != -1 else start
        document._cursors.add(self)

    def __repr__(self):
        key = [self.start]
        if self.start != self.end:
            key.append(self.end or "")
        key = ":".join(map(format, key))
        text = self.text()
        if len(text) > 30:
            text = text[:28] + "..."
        return "<{} [{}] {}>".format(type(self).__name__, key, repr(text))

    def document(self):
        return self._document

    def text(self):
        """Return the selected text, if any."""
        return self._document[self]

    def select(self, start, end=-1):
        """Change start and end in one go. End defaults to start."""
        self.start = start
        self.end = start if end == -1 else end

    def select_all(self):
        """Set start to 0 and end to None; selecting all text."""
        self.start = 0
        self.end = None

    def select_none(self):
        """Set end to start."""
        self.end = self.start

    def has_selection(self):
        """Return True if text is selected."""
        end = len(self._document) if self.end is None else self.end
        return self.start < end

    def lstrip(self, chars=None):
        """Move start to the right, if specified characters can be skipped.

        By default whitespace is skipped, like Python's lstrip() string method.

        """
        text = self.text()
        if text:
            offset = len(text) - len(text.lstrip(chars))
            self.start += offset

    def rstrip(self, chars=None):
        """Move end to the left, if specified characters can be skipped.

        By default whitespace is skipped, like Python's rstrip() string method.

        """
        text = self.text()
        if text:
            offset = len(text) - len(text.rstrip(chars))
            if offset:
                if self.end is None or self.end > len(self._document):
                    self.end = len(self._document)
                self.end -= offset

    def strip(self, chars=None):
        """Adjust start and end, like Python's strip() method."""
        self.rstrip(chars)
        self.lstrip(chars)





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Parser for LilyPond syntax.
"""


import re


from livelex import *


RE_LILYPOND_ID_RIGHT_BOUND = r"(?![_-]?[^\W\d])"
RE_LILYPOND_ID = r"[^\W\d_]+(?:[_-][^\W\d_]+)*"
RE_LILYPOND_VARIABLE = RE_LILYPOND_ID + RE_LILYPOND_ID_RIGHT_BOUND
RE_LILYPOND_COMMAND = r"\\(" + RE_LILYPOND_ID + ")" + RE_LILYPOND_ID_RIGHT_BOUND
RE_LILYPOND_MARKUP_TEXT = r'[^{}"\\\s#%]+'


class LilyPond(Language):

    @lexicon
    def root(cls):
        """Toplevel LilyPond document."""
        yield from cls.common()
        yield from cls.blocks()
        yield RE_LILYPOND_VARIABLE, Variable, cls.varname
        yield "=", Operator.Assignment
        yield r"\\version\b", Keyword

    @classmethod
    def blocks(cls):
        yield r"(\\book(?:part)?)\s*(\{)", bygroup(Keyword, Delimiter.OpenBrace), cls.book
        yield r"(\\score)\s*(\{)", bygroup(Keyword, Delimiter.OpenBrace), cls.score
        yield r"(\\header)\s*(\{)", bygroup(Keyword, Delimiter.OpenBrace), cls.header
        yield r"(\\paper)\s*(\{)", bygroup(Keyword, Delimiter.OpenBrace), cls.paper
        yield r"(\\layout)\s*(\{)", bygroup(Keyword, Delimiter.OpenBrace), cls.layout
        yield r"(\\midi)\s*(\{)", bygroup(Keyword, Delimiter.OpenBrace), cls.midi

    @lexicon
    def book(cls):
        """Book or bookpart."""
        yield r'\}', Delimiter.CloseBrace, -1
        yield from cls.common()
        yield from cls.blocks()

    @lexicon
    def score(cls):
        yield r'\}', Delimiter.CloseBrace, -1
        yield from cls.blocks()

    @lexicon
    def header(cls):
        yield r'\}', Delimiter.CloseBrace, -1
        yield RE_LILYPOND_VARIABLE, Variable, cls.varname
        yield "=", Operator.Assignment
        yield from cls.common()

    @lexicon
    def paper(cls):
        yield r'\}', Delimiter.CloseBrace, -1
        yield from cls.base()

    @lexicon
    def layout(cls):
        yield r'\}', Delimiter.CloseBrace, -1
        yield from cls.base()

    @lexicon
    def midi(cls):
        yield r'\}', Delimiter.CloseBrace, -1
        yield from cls.base()


    @classmethod
    def base(cls):
        """Find comment, string and scheme."""
        yield r'%\{', Comment, cls.multiline_comment
        yield r'%', Comment, cls.singleline_comment
        yield r'"', String, cls.string
        yield r'[#$]', Delimiter.SchemeStart, cls.scheme

    @classmethod
    def common(cls):
        """Find comment, string, scheme and markup."""
        yield from cls.base()
        yield r"\\markup(lines|list)?" + RE_LILYPOND_ID_RIGHT_BOUND, Keyword.Markup, cls.markup

    @lexicon
    def varname(cls):
        """bla.bla.bla syntax."""
        yield r'\s*(\.)\s*(' + RE_LILYPOND_VARIABLE + ')', bygroup(Delimiter.Dot, Variable)
        yield default_target, -1
        
    # -------------------- markup --------------------
    @lexicon
    def markup(cls):
        """Markup without environment. Try to guess the n of arguments."""
        yield r'\{', Delimiter.OpenBrace, -1, cls.markup_environ
        yield RE_LILYPOND_COMMAND, cls.get_markup_action(), cls.get_markup_target()
        yield r'"', String, -1, cls.string
        yield r'[#$]', Delimiter.SchemeStart, -1, cls.scheme
        yield r'%\{', Comment, cls.multiline_comment
        yield r'%', Comment, cls.singleline_comment
        yield RE_LILYPOND_MARKUP_TEXT, Text, -1

    @classmethod
    def get_markup_target(cls):
        """Get the target for a markup command."""
        def test(m):
            command = m.group(m.lastindex + 1)
            return cls.get_markup_argument_count(command)
        return tomatch(test, -1, 0, 1, 2, 3)

    @classmethod
    def get_markup_argument_count(cls, command):
        """Return the number of arguments the markup command (without \\) expects."""
        from . import lilypond_words
        for i in range(5):
            if command in lilypond_words.markupcommands_nargs[i]:
                return i
        return 1    # assume a user command has no arguments

    @lexicon
    def markup_environ(cls):
        """Markup until } ."""
        yield r'\}', Delimiter.CloseBrace, -1
        yield r'\{', Delimiter.OpenBrace, 1
        yield RE_LILYPOND_COMMAND, cls.get_markup_action()
        yield from cls.base()
        yield RE_LILYPOND_MARKUP_TEXT, Text

    @classmethod
    def get_markup_action(cls):
        """Get the action for a command in \markup { }."""
        def test(m):
            from . import lilypond_words
            text = m.group(m.lastindex + 1)
            return text in lilypond_words.markupcommands
        return bymatch(test, Function, Function.Markup)


    # -------------- Scheme ---------------------
    @lexicon
    def scheme(cls):
        from .scheme import SchemeLily
        yield from SchemeLily.one_arg()
        yield default_target, -1
        
    @lexicon
    def schemelily(cls):
        """LilyPond from scheme.SchemeLily #{ #}."""
        yield r"#}", Delimiter.LilyPond, -1
        yield from cls.root

    # -------------- String ---------------------
    @lexicon
    def string(cls):
        yield r'"', String, -1
        yield from cls.string_common()

    @classmethod
    def string_common(cls):
        yield r'\\[\\"]', String.Escape
        yield default_action, String

    # -------------- Comment ---------------------
    @lexicon
    def multiline_comment(cls):
        yield r'%}', Comment, -1
        yield from cls.comment_common()

    @lexicon(re_flags=re.MULTILINE)
    def singleline_comment(cls):
        yield from cls.comment_common()
        yield r'$', Comment, -1

    @classmethod
    def comment_common(cls):
        yield default_action, Comment





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Scheme keywords list.
Currently taken from ly.lex.scheme, but should be generated somehow.
"""

keywords = [
    '*',
    '+',
    '-',
    '->char-set',
    '/',
    '1+',
    '1-',
    '<',
    '<=',
    '=',
    '==',
    '>',
    '>=',
    'abs',
    'access?',
    'acons',
    'acos',
    'acosh',
    'activate-readline',
    'add-duration',
    'add-duration!',
    'add-hook!',
    'alist->hash-table',
    'alist-cons',
    'alist-copy',
    'alist-delete',
    'alist-delete!',
    'and',
    'and-let*',
    'angle',
    'any',
    'any->c32vector',
    'any->c64vector',
    'any->f32vector',
    'any->f64vector',
    'any->s16vector',
    'any->s32vector',
    'any->s64vector',
    'any->s8vector',
    'any->u16vector',
    'any->u32vector',
    'any->u64vector',
    'any->u8vector',
    'any-bits-set?',
    'append',
    'append!',
    'append-map',
    'append-map!',
    'append-reverse',
    'append-reverse!',
    'apply',
    'apply:nconc2last',
    'apropos-completion-function',
    'arithmetic-shift',
    'array->list',
    'array-contents',
    'array-copy!',
    'array-copy-in-order!',
    'array-dimensions',
    'array-equal?',
    'array-fill!',
    'array-for-each',
    'array-in-bounds?',
    'array-index-map!',
    'array-map!',
    'array-map-in-order!',
    'array-rank',
    'array-ref',
    'array-set!',
    'array-shape',
    'array-type',
    'array?',
    'ash',
    'asin',
    'asinh',
    'assoc',
    'assoc-ref',
    'assoc-remove!',
    'assoc-set!',
    'assq',
    'assq-ref',
    'assq-remove!',
    'assq-set!',
    'assv',
    'assv-ref',
    'assv-remove!',
    'assv-set!',
    'atan',
    'atanh',
    'basename',
    'begin',
    'bit-count',
    'bit-count*',
    'bit-extract',
    'bit-field',
    'bit-invert!',
    'bit-position',
    'bit-set*!',
    'bit-set?',
    'bitvector',
    'bitvector->list',
    'bitvector-fill!',
    'bitvector-length',
    'bitvector-ref',
    'bitvector-set!',
    'bitvector?',
    'bitwise-and',
    'bitwise-if',
    'bitwise-ior',
    'bitwise-merge',
    'bitwise-not',
    'bitwise-xor',
    'boolean?',
    'booleans->integer',
    'break',
    'break!',
    'broadcast-condition-variable',
    'c32vector',
    'c32vector->list',
    'c32vector-length',
    'c32vector-ref',
    'c32vector-set!',
    'c32vector?',
    'c64vector',
    'c64vector->list',
    'c64vector-length',
    'c64vector-ref',
    'c64vector-set!',
    'c64vector?',
    'caaaar',
    'caaadr',
    'caaar',
    'caadar',
    'caaddr',
    'caadr',
    'caar',
    'cadaar',
    'cadadr',
    'cadar',
    'caddar',
    'cadddr',
    'caddr',
    'cadr',
    'call-with-input-file',
    'call-with-input-string',
    'call-with-output-file',
    'call-with-output-string',
    'call-with-values',
    'car',
    'car+cdr',
    'case',
    'case-lambda',
    'catch',
    'cd',
    'cdaaar',
    'cdaadr',
    'cdaar',
    'cdadar',
    'cdaddr',
    'cdadr',
    'cdar',
    'cddaar',
    'cddadr',
    'cddar',
    'cdddar',
    'cddddr',
    'cdddr',
    'cddr',
    'cdr',
    'ceiling',
    'char->integer',
    'char-alphabetic?',
    'char-ci<=?',
    'char-ci<?',
    'char-ci=?',
    'char-ci>=?',
    'char-ci>?',
    'char-downcase',
    'char-is-both?',
    'char-lower-case?',
    'char-numeric?',
    'char-ready?',
    'char-set',
    'char-set->list',
    'char-set->string',
    'char-set-adjoin',
    'char-set-adjoin!',
    'char-set-any',
    'char-set-complement',
    'char-set-complement!',
    'char-set-contains?',
    'char-set-copy',
    'char-set-count',
    'char-set-cursor',
    'char-set-cursor-next',
    'char-set-delete',
    'char-set-delete!',
    'char-set-diff+intersection',
    'char-set-diff+intersection!',
    'char-set-difference',
    'char-set-difference!',
    'char-set-every',
    'char-set-filter',
    'char-set-filter!',
    'char-set-fold',
    'char-set-for-each',
    'char-set-hash',
    'char-set-intersection',
    'char-set-intersection!',
    'char-set-map',
    'char-set-ref',
    'char-set-size',
    'char-set-unfold',
    'char-set-unfold!',
    'char-set-union',
    'char-set-union!',
    'char-set-xor',
    'char-set-xor!',
    'char-set<=',
    'char-set=',
    'char-set?',
    'char-upcase',
    'char-upper-case?',
    'char-whitespace?',
    'char<=?',
    'char<?',
    'char=?',
    'char>=?',
    'char>?',
    'char?',
    'chdir',
    'chmod',
    'chown',
    'chroot',
    'circular-list',
    'circular-list?',
    'close',
    'close-fdes',
    'close-input-port',
    'close-output-port',
    'close-pipe',
    'close-port',
    'closedir',
    'closure?',
    'command-line',
    'complex?',
    'concatenate',
    'concatenate!',
    'cond',
    'cond-expand',
    'condition',
    'condition-has-type?',
    'condition-message',
    'condition-ref',
    'condition-type?',
    'cons',
    'cons*',
    'cons-source',
    'continue',
    'copy-bit',
    'copy-bit-field',
    'copy-file',
    'copy-random-state',
    'copy-time',
    'copy-tree',
    'cos',
    'cosh',
    'count',
    'ctermid',
    'current-date',
    'current-dynamic-state',
    'current-error-port',
    'current-input-port',
    'current-julian-day',
    'current-load-port',
    'current-modified-julian-day',
    'current-module',
    'current-output-port',
    'current-time',
    'cut',
    'cute',
    'date->julian-day',
    'date->modified-julian-day',
    'date->string',
    'date->time-monotonic',
    'date->time-tai',
    'date->time-utc',
    'date-day',
    'date-hour',
    'date-minute',
    'date-month',
    'date-nanosecond',
    'date-second',
    'date-week-day',
    'date-week-number',
    'date-year',
    'date-year-day',
    'date-zone-offset',
    'date?',
    'debug-object?',
    'debug-options-interface',
    'debug-trap',
    'default-duplicate-binding-handler',
    'define',
    'define*',
    'define*-public',
    'define-condition-type',
    'define-macro',
    'define-module',
    'define-public',
    'define-reader-ctor',
    'define-record-type',
    'defined?',
    'defmacro',
    'defmacro*',
    'defmacro*-public',
    'delay',
    'delete',
    'delete!',
    'delete-duplicates',
    'delete-duplicates!',
    'delete-file',
    'delete1!',
    'delq',
    'delq!',
    'delq1!',
    'delv',
    'delv!',
    'delv1!',
    'denominator',
    'deq!',
    'directory-stream?',
    'dirname',
    'display',
    'display-application',
    'display-backtrace',
    'display-error',
    'do',
    'dotted-list?',
    'doubly-weak-hash-table?',
    'down',
    'drain-input',
    'drop',
    'drop-right',
    'drop-right!',
    'drop-while',
    'dup',
    'dup->fdes',
    'dup->inport',
    'dup->outport',
    'dup->port',
    'dup2',
    'duplicate-port',
    'dynamic-args-call',
    'dynamic-call',
    'dynamic-func',
    'dynamic-link',
    'dynamic-object?',
    'dynamic-state?',
    'dynamic-unlink',
    'dynamic-wind',
    'effective-version',
    'eighth',
    'else',
    'enclose-array',
    'end-of-char-set?',
    'enq!',
    'entity?',
    'environ',
    'eof-object?',
    'eq?',
    'equal?',
    'eqv?',
    'error',
    'error?',
    'eval',
    'eval-disable',
    'eval-enable',
    'eval-options',
    'eval-options-interface',
    'eval-set!',
    'eval-string',
    'evaluate',
    'evaluator-traps-interface',
    'even?',
    'every',
    'exact->inexact',
    'exact?',
    'execl',
    'execle',
    'execlp',
    'exp',
    'expect',
    'expect-strings',
    'export',
    'expt',
    'extract-condition',
    'f32vector',
    'f32vector->list',
    'f32vector-length',
    'f32vector-ref',
    'f32vector-set!',
    'f32vector?',
    'f64vector',
    'f64vector->list',
    'f64vector-length',
    'f64vector-ref',
    'f64vector-set!',
    'f64vector?',
    'false-if-exception',
    'fchmod',
    'fchown',
    'fcntl',
    'fdes->inport',
    'fdes->outport',
    'fdes->ports',
    'fdopen',
    'feature?',
    'fflush',
    'fifth',
    'file-exists?',
    'file-port?',
    'filename-completion-function',
    'fileno',
    'filter',
    'filter!',
    'filter-map',
    'find',
    'find-tail',
    'first',
    'first-set-bit',
    'flock',
    'floor',
    'fluid-ref',
    'fluid-set!',
    'fluid?',
    'flush-all-ports',
    'fn',
    'fold',
    'fold-matches',
    'fold-right',
    'for-each',
    'force',
    'force-output',
    'format',
    'fourth',
    'frame',
    'frame-arguments',
    'frame-evaluating-args?',
    'frame-next',
    'frame-number',
    'frame-overflow?',
    'frame-previous',
    'frame-procedure',
    'frame-procedure?',
    'frame-real?',
    'frame-source',
    'frame?',
    'fstat',
    'fsync',
    'ftell',
    'ftruncate',
    'ftw',
    'gcd',
    'generalized-vector->list',
    'generalized-vector-length',
    'generalized-vector-ref',
    'generalized-vector-set!',
    'generalized-vector?',
    'gensym',
    'get-internal-real-time',
    'get-internal-run-time',
    'get-output-string',
    'get-print-state',
    'getcwd',
    'getegid',
    'getenv',
    'geteuid',
    'getgid',
    'getgroups',
    'getpgrp',
    'getpid',
    'getppid',
    'getpriority',
    'getter-with-setter',
    'gettimeofday',
    'getuid',
    'gmtime',
    'hash',
    'hash-by-identity',
    'hash-clear!',
    'hash-create-handle!',
    'hash-fold',
    'hash-for-each',
    'hash-for-each-handle',
    'hash-get-handle',
    'hash-map->list',
    'hash-ref',
    'hash-remove!',
    'hash-set!',
    'hash-table->alist',
    'hash-table-delete!',
    'hash-table-equivalence-function',
    'hash-table-exists?',
    'hash-table-fold',
    'hash-table-hash-function',
    'hash-table-keys',
    'hash-table-ref',
    'hash-table-ref/default',
    'hash-table-set!',
    'hash-table-size',
    'hash-table-update!',
    'hash-table-update!/default',
    'hash-table-values',
    'hash-table-walk',
    'hash-table?',
    'hashq',
    'hashq-create-handle!',
    'hashq-get-handle',
    'hashq-ref',
    'hashq-remove!',
    'hashq-set!',
    'hashv',
    'hashv-create-handle!',
    'hashv-get-handle',
    'hashv-ref',
    'hashv-remove!',
    'hashv-set!',
    'hashx-create-handle!',
    'hashx-get-handle',
    'hashx-ref',
    'hashx-remove!',
    'hashx-set!',
    'hook->list',
    'hook-empty?',
    'hook?',
    'if',
    'imag-part',
    'inet-aton',
    'inet-lnaof',
    'inet-makeaddr',
    'inet-netof',
    'inet-ntoa',
    'inet-ntop',
    'inet-pton',
    'inexact->exact',
    'inexact?',
    'inf',
    'inf?',
    'input-port?',
    'install-trap',
    'integer->char',
    'integer->list',
    'integer-expt',
    'integer-length',
    'integer?',
    'interaction-environment',
    'iota',
    'isatty?',
    'julian-day->date',
    'julian-day->time-monotonic',
    'julian-day->time-tai',
    'julian-day->time-utc',
    'key',
    'keyword->string',
    'keyword->symbol',
    'keyword?',
    'lambda',
    'lambda*',
    'last',
    'last-pair',
    'last-stack-frame',
    'lazy-catch',
    'lchown',
    'lcm',
    'length',
    'length+',
    'let',
    'let*',
    'let*-values',
    'let-keywords',
    'let-keywords*',
    'let-optional',
    'let-optional*',
    'let-values',
    'letpar',
    'letrec',
    'link',
    'list',
    'list->array',
    'list->bitvector',
    'list->c32vector',
    'list->c64vector',
    'list->char-set',
    'list->char-set!',
    'list->f32vector',
    'list->f64vector',
    'list->integer',
    'list->s16vector',
    'list->s32vector',
    'list->s64vector',
    'list->s8vector',
    'list->stream',
    'list->string',
    'list->typed-array',
    'list->u16vector',
    'list->u32vector',
    'list->u64vector',
    'list->u8vector',
    'list->vector',
    'list->weak-vector',
    'list-cdr-ref',
    'list-cdr-set!',
    'list-copy',
    'list-head',
    'list-index',
    'list-matches',
    'list-ref',
    'list-set!',
    'list-tabulate',
    'list-tail',
    'list=',
    'list?',
    'load',
    'load-extension',
    'load-from-path',
    'local-eval',
    'localtime',
    'lock-mutex',
    'log',
    'log10',
    'log2-binary-factors',
    'logand',
    'logbit?',
    'logcount',
    'logior',
    'lognot',
    'logtest',
    'logxor',
    'lset-adjoin',
    'lset-diff+intersection',
    'lset-diff+intersection!',
    'lset-difference',
    'lset-difference!',
    'lset-intersection',
    'lset-intersection!',
    'lset-union',
    'lset-union!',
    'lset-xor',
    'lset-xor!',
    'lset<=',
    'lset=',
    'lstat',
    'macro-name',
    'macro-transformer',
    'macro-type',
    'macro?',
    'magnitude',
    'major-version',
    'make-arbiter',
    'make-array',
    'make-bitvector',
    'make-buffered-input-port',
    'make-c32vector',
    'make-c64vector',
    'make-class-object',
    'make-completion-function',
    'make-compound-condition',
    'make-condition',
    'make-condition-type',
    'make-condition-variable',
    'make-date',
    'make-doubly-weak-hash-table',
    'make-dynamic-state',
    'make-f32vector',
    'make-f64vector',
    'make-fluid',
    'make-guardian',
    'make-hash-table',
    'make-hook',
    'make-line-buffered-input-port',
    'make-list',
    'make-mutex',
    'make-object-property',
    'make-parameter',
    'make-polar',
    'make-procedure-with-setter',
    'make-q',
    'make-record-type',
    'make-rectangular',
    'make-recursive-mutex',
    'make-regexp',
    'make-s16vector',
    'make-s32vector',
    'make-s64vector',
    'make-s8vector',
    'make-shared-array',
    'make-soft-port',
    'make-stack',
    'make-stream',
    'make-string',
    'make-struct',
    'make-struct-layout',
    'make-subclass-object',
    'make-symbol',
    'make-time',
    'make-typed-array',
    'make-u16vector',
    'make-u32vector',
    'make-u64vector',
    'make-u8vector',
    'make-undefined-variable',
    'make-variable',
    'make-vector',
    'make-vtable',
    'make-vtable-vtable',
    'make-weak-key-hash-table',
    'make-weak-value-hash-table',
    'make-weak-vector',
    'malloc-stats',
    'map',
    'map!',
    'map-in-order',
    'match:count',
    'match:end',
    'match:prefix',
    'match:start',
    'match:string',
    'match:substring',
    'match:suffix',
    'max',
    'member',
    'memoized-environment',
    'memoized?',
    'memq',
    'memv',
    'merge',
    'merge!',
    'message-condition?',
    'micro-version',
    'min',
    'minor-version',
    'mkdir',
    'mknod',
    'mkstemp!',
    'mktime',
    'modified-julian-day->date',
    'modified-julian-day->time-monotonic',
    'modified-julian-day->time-tai',
    'modified-julian-day->time-utc',
    'module-use!',
    'modulo',
    'modulo-expt',
    'monitor',
    'move->fdes',
    'n-for-each-par-map',
    'n-par-for-each',
    'n-par-map',
    'nan',
    'nan?',
    'negative?',
    'newline',
    'nftw',
    'nice',
    'nil-car',
    'nil-cdr',
    'nil-cons',
    'nil-eq',
    'ninth',
    'not',
    'not-pair?',
    'null',
    'null-environment',
    'null-list?',
    'null?',
    'number->string',
    'number?',
    'numerator',
    'object->string',
    'object-properties',
    'object-property',
    'odd?',
    'open',
    'open-fdes',
    'open-file',
    'open-input-file',
    'open-input-output-pipe',
    'open-input-pipe',
    'open-input-string',
    'open-output-file',
    'open-output-pipe',
    'open-output-string',
    'open-pipe',
    'open-pipe*',
    'opendir',
    'operator?',
    'option-ref',
    'or',
    'output-port?',
    'pair-fold',
    'pair-fold-right',
    'pair-for-each',
    'pair?',
    'par-for-each',
    'par-map',
    'parallel',
    'parameterize',
    'parse-path',
    'partition',
    'partition!',
    'pclose',
    'peek-char',
    'pipe',
    'popen',
    'port->fdes',
    'port->stream',
    'port-closed?',
    'port-column',
    'port-filename',
    'port-for-each',
    'port-line',
    'port-mode',
    'port-revealed',
    'port-with-print-state',
    'port?',
    'position',
    'positive?',
    'pretty-print',
    'primitive-eval',
    'primitive-exit',
    'primitive-fork',
    'primitive-load',
    'primitive-load-path',
    'primitive-make-property',
    'primitive-move->fdes',
    'primitive-property-del!',
    'primitive-property-ref',
    'primitive-property-set!',
    'print-options-interface',
    'procedure',
    'procedure->macro',
    'procedure->memoizing-macro',
    'procedure->syntax',
    'procedure-documentation',
    'procedure-environment',
    'procedure-name',
    'procedure-properties',
    'procedure-property',
    'procedure-source',
    'procedure-with-setter?',
    'procedure?',
    'program-arguments',
    'promise?',
    'proper-list?',
    'provide',
    'provided?',
    'putenv',
    'pwd',
    'q-empty-check',
    'q-empty?',
    'q-front',
    'q-length',
    'q-pop!',
    'q-push!',
    'q-rear',
    'q-remove!',
    'q?',
    'quasiquote',
    'quit',
    'quote',
    'quotient',
    'random',
    'random:exp',
    'random:hollow-sphere!',
    'random:normal',
    'random:normal-vector!',
    'random:solid-sphere!',
    'random:uniform',
    'rational?',
    'rationalize',
    're-export',
    'read',
    'read-char',
    'read-delimited',
    'read-delimited!',
    'read-disable',
    'read-enable',
    'read-hash-extend',
    'read-line',
    'read-line!',
    'read-options',
    'read-options-interface',
    'read-set!',
    'read-string!/partial',
    'readdir',
    'readline',
    'readline-disable',
    'readline-enable',
    'readline-options',
    'readline-port',
    'readline-set!',
    'readlink',
    'real-part',
    'real?',
    'rec',
    'receive',
    'record-accessor',
    'record-constructor',
    'record-modifier',
    'record-predicate',
    'record-type-descriptor',
    'record-type-fields',
    'record-type-name',
    'record?',
    'redirect-port',
    'reduce',
    'reduce-right',
    'regexp-exec',
    'regexp-match?',
    'regexp-quote',
    'regexp-substitute',
    'regexp-substitute/global',
    'regexp?',
    'release-arbiter',
    'release-port-handle',
    'remainder',
    'remove',
    'remove!',
    'remove-hook!',
    'rename',
    'rename-file',
    'require',
    'require-extension',
    'reset-hook!',
    'resolve-interface',
    'resolve-module',
    'restricted-vector-sort!',
    'reverse',
    'reverse!',
    'reverse-bit-field',
    'reverse-list->string',
    'rewinddir',
    'rmdir',
    'rotate-bit-field',
    'round',
    'run-hook',
    's16vector',
    's16vector->list',
    's16vector-length',
    's16vector-ref',
    's16vector-set!',
    's16vector?',
    's32vector',
    's32vector->list',
    's32vector-length',
    's32vector-ref',
    's32vector-set!',
    's32vector?',
    's64vector',
    's64vector->list',
    's64vector-length',
    's64vector-ref',
    's64vector-set!',
    's64vector?',
    's8vector',
    's8vector->list',
    's8vector-length',
    's8vector-ref',
    's8vector-set!',
    's8vector?',
    'save-module-excursion',
    'scheme-report-environment',
    'scm-error',
    'search-path',
    'second',
    'seed->random-state',
    'seek',
    'select',
    'serious-condition?',
    'set!',
    'set-buffered-input-continuation?!',
    'set-car!',
    'set-cdr!',
    'set-current-dynamic-state',
    'set-current-error-port',
    'set-current-input-port',
    'set-current-module',
    'set-current-output-port',
    'set-object-procedure!',
    'set-object-properties!',
    'set-object-property!',
    'set-port-column!',
    'set-port-filename!',
    'set-port-line!',
    'set-port-revealed!',
    'set-procedure-properties!',
    'set-procedure-property!',
    'set-program-arguments',
    'set-readline-input-port!',
    'set-readline-output-port!',
    'set-readline-prompt!',
    'set-source-properties!',
    'set-source-property!',
    'set-struct-vtable-name!',
    'set-symbol-property!',
    'set-time-nanosecond!',
    'set-time-second!',
    'set-time-type!',
    'set-tm:gmtoff',
    'set-tm:hour',
    'set-tm:isdst',
    'set-tm:mday',
    'set-tm:min',
    'set-tm:mon',
    'set-tm:sec',
    'set-tm:wday',
    'set-tm:yday',
    'set-tm:year',
    'set-tm:zone',
    'set-trace-layout',
    'setegid',
    'setenv',
    'seteuid',
    'setgid',
    'setgroups',
    'setlocale',
    'setpgid',
    'setpriority',
    'setsid',
    'setter',
    'setuid',
    'setvbuf',
    'seventh',
    'shared-array-increments',
    'shared-array-offset',
    'shared-array-root',
    'signal-condition-variable',
    'simple-format',
    'sin',
    'sinh',
    'sixth',
    'sloppy-assoc',
    'sloppy-assq',
    'sloppy-assv',
    'sort',
    'sort!',
    'sort-list',
    'sort-list!',
    'sorted?',
    'source-properties',
    'source-property',
    'span',
    'span!',
    'split-at',
    'split-at!',
    'sqrt',
    'stable-sort',
    'stable-sort!',
    'stack-id',
    'stack-length',
    'stack-ref',
    'stack?',
    'start-stack',
    'stat',
    'stat:atime',
    'stat:blksize',
    'stat:blocks',
    'stat:ctime',
    'stat:dev',
    'stat:gid',
    'stat:ino',
    'stat:mode',
    'stat:mtime',
    'stat:nlink',
    'stat:perms',
    'stat:rdev',
    'stat:size',
    'stat:type',
    'stat:uid',
    'status:exit-val',
    'status:stop-sig',
    'status:term-sig',
    'stream->list',
    'stream->list&amp;length',
    'stream->reversed-list',
    'stream->reversed-list&amp;length',
    'stream->vector',
    'stream-car',
    'stream-cdr',
    'stream-fold',
    'stream-for-each',
    'stream-map',
    'stream-null?',
    'strerror',
    'strftime',
    'string',
    'string->char-set',
    'string->char-set!',
    'string->date',
    'string->keyword',
    'string->list',
    'string->number',
    'string->symbol',
    'string-any',
    'string-append',
    'string-append/shared',
    'string-capitalize',
    'string-capitalize!',
    'string-ci->symbol',
    'string-ci-hash',
    'string-ci<',
    'string-ci<=',
    'string-ci<=?',
    'string-ci<>',
    'string-ci<?',
    'string-ci=',
    'string-ci=?',
    'string-ci>',
    'string-ci>=',
    'string-ci>=?',
    'string-ci>?',
    'string-compare',
    'string-compare-ci',
    'string-concatenate',
    'string-concatenate-reverse',
    'string-concatenate-reverse/shared',
    'string-concatenate/shared',
    'string-contains',
    'string-contains-ci',
    'string-copy',
    'string-copy!',
    'string-count',
    'string-delete',
    'string-downcase',
    'string-downcase!',
    'string-drop',
    'string-drop-right',
    'string-every',
    'string-fill!',
    'string-filter',
    'string-fold',
    'string-fold-right',
    'string-for-each',
    'string-for-each-index',
    'string-hash',
    'string-hash-ci',
    'string-index',
    'string-index-right',
    'string-join',
    'string-length',
    'string-map',
    'string-map!',
    'string-match',
    'string-null?',
    'string-pad',
    'string-pad-right',
    'string-prefix-ci?',
    'string-prefix-length',
    'string-prefix-length-ci',
    'string-prefix?',
    'string-ref',
    'string-replace',
    'string-reverse',
    'string-reverse!',
    'string-rindex',
    'string-set!',
    'string-skip',
    'string-skip-right',
    'string-split',
    'string-suffix-ci?',
    'string-suffix-length',
    'string-suffix-length-ci',
    'string-suffix?',
    'string-tabulate',
    'string-take',
    'string-take-right',
    'string-titlecase',
    'string-titlecase!',
    'string-tokenize',
    'string-trim',
    'string-trim-both',
    'string-trim-right',
    'string-unfold',
    'string-unfold-right',
    'string-upcase',
    'string-upcase!',
    'string-xcopy!',
    'string<',
    'string<=',
    'string<=?',
    'string<>',
    'string<?',
    'string=',
    'string=?',
    'string>',
    'string>=',
    'string>=?',
    'string>?',
    'string?',
    'strptime',
    'struct-ref',
    'struct-set!',
    'struct-vtable',
    'struct-vtable-name',
    'struct-vtable-tag',
    'struct-vtable?',
    'struct?',
    'substring',
    'substring-fill!',
    'substring-move!',
    'substring/copy',
    'substring/read-only',
    'substring/shared',
    'subtract-duration',
    'subtract-duration!',
    'symbol->keyword',
    'symbol->string',
    'symbol-fref',
    'symbol-fset!',
    'symbol-hash',
    'symbol-interned?',
    'symbol-pref',
    'symbol-prefix-proc',
    'symbol-property',
    'symbol-property-remove!',
    'symbol-pset!',
    'symbol?',
    'symlink',
    'sync',
    'sync-q!',
    'system',
    'system*',
    'system-error-errno',
    'take',
    'take!',
    'take-right',
    'take-while',
    'take-while!',
    'tan',
    'tanh',
    'tc:depth',
    'tc:frame',
    'tc:real-depth',
    'tc:return-value',
    'tc:stack',
    'tc:type',
    'tcgetpgrp',
    'tcsetpgrp',
    'tenth',
    'third',
    'throw',
    'thunk?',
    'time-difference',
    'time-difference!',
    'time-monotonic->date',
    'time-monotonic->time-tai',
    'time-monotonic->time-tai!',
    'time-monotonic->time-utc',
    'time-monotonic->time-utc!',
    'time-nanosecond',
    'time-resolution',
    'time-second',
    'time-tai->date',
    'time-tai->julian-day',
    'time-tai->modified-julian-day',
    'time-tai->time-monotonic',
    'time-tai->time-monotonic!',
    'time-tai->time-utc',
    'time-tai->time-utc!',
    'time-type',
    'time-utc->date',
    'time-utc->julian-day',
    'time-utc->modified-julian-day',
    'time-utc->time-monotonic',
    'time-utc->time-monotonic!',
    'time-utc->time-tai',
    'time-utc->time-tai!',
    'time<=?',
    'time<?',
    'time=?',
    'time>=?',
    'time>?',
    'time?',
    'times',
    'tm:gmtoff',
    'tm:hour',
    'tm:isdst',
    'tm:mday',
    'tm:min',
    'tm:mon',
    'tm:sec',
    'tm:wday',
    'tm:yday',
    'tm:year',
    'tm:zone',
    'tmpnam',
    'tms:clock',
    'tms:cstime',
    'tms:cutime',
    'tms:stime',
    'tms:utime',
    'trace',
    'trace-at-exit',
    'trace-port',
    'trace-trap',
    'trace-until-exit',
    'trace/info',
    'trace/pid',
    'trace/real?',
    'trace/source',
    'trace/source-column',
    'trace/source-file-name',
    'trace/source-line',
    'trace/stack',
    'trace/stack-depth',
    'trace/stack-id',
    'trace/stack-real-depth',
    'trace/type',
    'transpose-array',
    'trap-disable',
    'trap-enable',
    'trap-set!',
    'traps',
    'truncate',
    'truncate-file',
    'try-arbiter',
    'try-mutex',
    'ttyname',
    'typed-array?',
    'tzset',
    'u16vector',
    'u16vector->list',
    'u16vector-length',
    'u16vector-ref',
    'u16vector-set!',
    'u16vector?',
    'u32vector',
    'u32vector->list',
    'u32vector-length',
    'u32vector-ref',
    'u32vector-set!',
    'u32vector?',
    'u64vector',
    'u64vector->list',
    'u64vector-length',
    'u64vector-ref',
    'u64vector-set!',
    'u64vector?',
    'u8vector',
    'u8vector->list',
    'u8vector-length',
    'u8vector-ref',
    'u8vector-set!',
    'u8vector?',
    'ucs-range->char-set',
    'ucs-range->char-set!',
    'umask',
    'unfold',
    'unfold-right',
    'uniform-array-read!',
    'uniform-array-write',
    'uniform-vector->list',
    'uniform-vector-length',
    'uniform-vector-read!',
    'uniform-vector-ref',
    'uniform-vector-set!',
    'uniform-vector-write',
    'uniform-vector?',
    'uninstall-trap',
    'unlink',
    'unlock-mutex',
    'unmemoize',
    'unquote',
    'unquote-splicing',
    'unread-char',
    'unread-string',
    'unsetenv',
    'untrace',
    'unzip1',
    'unzip2',
    'unzip3',
    'unzip4',
    'unzip5',
    'up',
    'use-modules',
    'use-syntax',
    'utime',
    'values',
    'variable-bound?',
    'variable-ref',
    'variable-set!',
    'variable?',
    'vector',
    'vector->list',
    'vector->stream',
    'vector-copy',
    'vector-fill!',
    'vector-length',
    'vector-move-left!',
    'vector-move-right!',
    'vector-ref',
    'vector-set!',
    'vector?',
    'version',
    'wait-condition-variable',
    'waitpid',
    'weak-key-hash-table?',
    'weak-value-hash-table?',
    'weak-vector',
    'weak-vector?',
    'while',
    'with-continuation-barrier',
    'with-dynamic-state',
    'with-error-to-file',
    'with-fluid*',
    'with-fluids',
    'with-fluids*',
    'with-input-from-file',
    'with-input-from-string',
    'with-mutex',
    'with-output-to-file',
    'with-output-to-string',
    'with-parameters*',
    'with-readline-completion-function',
    'with-throw-handler',
    'with-traps',
    'write',
    'write-char',
    'write-line',
    'write-string/partial',
    'xcons',
    'xsubstring',
    'zero?',
    'zip',
]



# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re


from livelex import *


CDATA = Literal.CDATA
DOCTYPE = Keyword.DOCTYPE
ENTITY = Keyword.ENTITY
PI = Comment.PI


class Xml(Language):
    """Parse XML."""
    @lexicon(re_flags=re.IGNORECASE)
    def root(cls):
        yield r'<!--', Comment.Start, cls.comment
        yield r'<!\[CDATA\[', CDATA.Start, cls.cdata
        yield r'<!DOCTYPE\b', DOCTYPE.Start, cls.doctype
        yield r'<\?', PI.Start, cls.pi
        yield r'(<\s*?/)\s*(\w+(?:[:.-]\w+)*)\s*(>)', bygroup(Delimiter, Name.Tag, Delimiter), -1
        yield r'(<)\s*(\w+(?:[:.-]\w+)*)\s*(>)', bygroup(Delimiter, Name.Tag, Delimiter), cls.tag
        yield r'(<)\s*(\w+(?:[:.-]\w+)*)', bygroup(Delimiter, Name.Tag), cls.attrs
        yield r'&\S*?;', Escape.Entity
        yield default_action, Text

    @lexicon
    def tag(cls):
        yield from cls.root()

    @lexicon
    def comment(cls):
        yield default_action, Comment
        yield r'-->', Comment.End, -1

    @lexicon
    def cdata(cls):
        yield default_action, CDATA
        yield r'\]\]>', CDATA.End, -1

    @lexicon
    def pi(cls):
        yield r'(\w+(?:[:.-]\w+)*)\s*?(=)(?=\s*?")', bygroup(Name.Attribute, Operator)
        yield r'"', String, cls.dqstring
        yield default_action, PI
        yield r'\?>', PI.End, -1

    @lexicon
    def doctype(cls):
        yield r'\w+', Text
        yield r'"', String, cls.dqstring
        yield r'\[', DOCTYPE.Start, cls.internal_dtd
        yield r'>', DOCTYPE.End, -1

    @lexicon
    def internal_dtd(cls):
        yield r'<!ENTITY\b', ENTITY.Start, cls.entity
        yield r'<![^>]*>', DOCTYPE
        yield default_action, Text  # TODO include dtd language
        yield r'\]', DOCTYPE.End, -1

    @lexicon
    def entity(cls):
        yield r'\w+', Name.Entity
        yield r'"', String, cls.dqstring
        yield r'>', ENTITY.End, -1

    @lexicon
    def attrs(cls):
        yield r'\w+([:.-]\w+)*', Name.Attribute
        yield r'=', Operator
        yield r'"', String, cls.dqstring
        yield r"'", String, cls.sqstring
        yield r'/\s*>', Delimiter, -1
        yield r'>', Delimiter, -1, cls.tag
        yield r'\s+', skip
        yield default_action, Error

    @lexicon
    def dqstring(cls):
        yield r'&\S*?;', Escape.Entity
        yield default_action, String
        yield r'"', String, -1

    @lexicon
    def sqstring(cls):
        yield r'&\S*?;', Escape.Entity
        yield default_action, String
        yield r"'", String, -1




# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Various utility functions.
"""

def abbreviate_repr(s, length=30):
    """Elegantly abbreviate repr text."""
    if len(s) > length:
        return repr(s[:length-2]) + "..."
    return repr(s)


def merge_adjacent_actions(tokens):
    """Yield three-tuples (pos, end, action).

    Adjacent actions that are the same are merged into
    one range.

    """
    stream = ((t.pos, t.end, t.action) for t in tokens)
    for pos, end, action in stream:
        for npos, nend, naction in stream:
            if naction != action or npos > end:
                yield pos, end, action
                pos, action = npos, naction
            end = nend
        yield pos, end, action





# -*- coding: utf-8 -*-
#
# This file is part of the livelex Python package.
#
# Copyright © 2019 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re

import livelex
from livelex.lexicon import Lexicon, BoundLexicon
from livelex.pattern import Pattern
from livelex.target import DynamicTarget


def validate_language(lang):
    """Perform checks to the specified language class.

    Detects circular default targets, invalid regular expressions, etc.

    """
    lexicons = []
    for key, value in lang.__dict__.items():
        if isinstance(value, Lexicon):
            lexicons.append(getattr(lang, key))

    for lexicon in lexicons:
        validate_lexicon(lexicon)


def validate_lexicon(lexicon):
    default_act, default_tg = None, None
    msg = message(lexicon)
    for pattern, action, *target in lexicon():
        if pattern is livelex.default_action:
            if default_act:
                msg("conflicting default actions")
            else:
                default_act = action
        elif pattern is livelex.default_target:
            if default_tg:
                msg("conflicting default targets")
            else:
                default_tg = action, *target
                check_default_target(lexicon, default_tg)
        else:
            validate_pattern(msg, pattern)
            validate_target(msg, target)

    if default_act and default_tg:
        msg("can't have both default_action and default_target")


def validate_pattern(msg, pattern):
    """Validate a regular expression pattern."""
    if isinstance(pattern, Pattern):
        pattern = pattern.build()
    try:
        rx = re.compile(pattern)
    except re.error as e:
        msg("regular expression {} error:\n  {}".format(repr(pattern), e))
    else:
        if rx.match(''):
            msg("warning: pattern {} matches the empty string".format(repr(pattern)))


def validate_target(msg, target):
    """Validate a target."""
    def targets():
        if len(target) == 1 and isinstance(target[0], DynamicTarget):
            for t in target[0].targets:
                yield from t
        else:
            yield from target
    for t in targets():
        if isinstance(t, DynamicTarget):
            msg("a DynamicTarget must be the only one: {}".format(target))
        elif not isinstance(t, (int, BoundLexicon)):
            msg("invalid target {} in targets {}".format(t, target))
            break


def check_default_target(lexicon, target):
    """Check whether this default target could lead to circular references.

    This could hang the parser, and we wouldn't like to have that :-)

    """
    state = [lexicon]
    circular = set()
    while True:
        msg = message(lexicon)
        circular.add(lexicon)
        depth = len(state)
        for t in target:
            if isinstance(t, int):
                if t < 0:
                    if len(state) + t < 1:
                        return
                    del state[t:]
                else:
                    state += [lexicon] * t
            elif not isinstance(t, BoundLexicon):
                msg("in default target only integer or lexicon allowed")
                return
            else:
                state.append(t)
        if len(state) == depth:
            msg("invalid default target")
            return
        lexicon = state[-1]
        if lexicon in circular:
            state.extend(l for l in circular if l not in state)
            msg("circular default target: {}".format(" -> ".join(map(str, state))))
            return
        for pattern, *target in lexicon():
            if pattern is livelex.default_target:
                break
        else:
            break


def message(lexicon):
    """Return a callable that prints a message with the lexicon name prepended."""
    return lambda s: print("{}: ".format(lexicon) + s)

